import json, os, sys
from os.path import join as opj
import numpy as np
import numpy
import requests
import xmltodict
import pandas as pd
import time
from numpy import array as npar
import sys
from scipy import interpolate
import traceback
from datetime import datetime
import datetime
import copy
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

THICKNESS = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]
hydro = {'A': 67, 'B': 78, 'C': 85, 'D': 89}

import requests
import xmltodict


def DownloadsurgoSoiltables(lonlat, select_componentname=None, summarytable=False):
    '''
    Downloads SSURGO soil tables

    Parameters
    ------------------
    :param lonlat: tuple of (longitude, latitude)
    :param select_componentname: specific component name within the map unit, default None
    :param summarytable: if True, prints summary table of component names and their percentages
    '''

    lonLat = f"{lonlat[0]} {lonlat[1]}"
    url = "https://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx"

    # FIXED: Correct header name and value
    headers = {
        'Content-Type': 'text/xml; charset=utf-8',
        'SOAPAction': 'http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx/RunQuery'
    }

    body = f"""<?xml version="1.0" encoding="utf-8"?>
    <soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:sdm="http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx">
       <soap:Header/>
       <soap:Body>
          <sdm:RunQuery>
             <sdm:Query>
             SELECT co.cokey as cokey, ch.chkey as chkey, comppct_r as prcent, compkind as compkind_series, 
             wsatiated_r as wat_r,partdensity as pd, dbthirdbar_h as bb, musym as musymbol, compname as componentname, 
             muname as muname, slope_r, slope_h as slope, hzname, hzdept_r as topdepth, hzdepb_r as bottomdepth, 
             awc_r as PAW, ksat_l as KSAT, claytotal_r as clay, silttotal_r as silt, sandtotal_r as sand, texcl, 
             drainagecl, om_r as OM, iacornsr as CSR, dbthirdbar_r as BD, wfifteenbar_r as L15, wthirdbar_h as DUL, 
             ph1to1h2o_r as pH, ksat_r as sat_hidric_cond, (dbthirdbar_r-wthirdbar_r)/100 as bd 
             FROM sacatalog sc
             FULL OUTER JOIN legend lg  ON sc.areasymbol=lg.areasymbol
             FULL OUTER JOIN mapunit mu ON lg.lkey=mu.lkey
             FULL OUTER JOIN component co ON mu.mukey=co.mukey
             FULL OUTER JOIN chorizon ch ON co.cokey=ch.cokey
             FULL OUTER JOIN chtexturegrp ctg ON ch.chkey=ctg.chkey
             FULL OUTER JOIN chtexture ct ON ctg.chtgkey=ct.chtgkey
             FULL OUTER JOIN copmgrp pmg ON co.cokey=pmg.cokey
             FULL OUTER JOIN corestrictions rt ON co.cokey=rt.cokey
             WHERE mu.mukey IN (
                 SELECT * from SDA_Get_Mukey_from_intersection_with_WktWgs84('point({lonLat})')
             ) 
             AND sc.areasymbol != 'US' 
             ORDER BY co.cokey, ch.chkey, prcent, topdepth, bottomdepth, muname
             </sdm:Query>
          </sdm:RunQuery>
       </soap:Body>
    </soap:Envelope>"""

    response = requests.post(url, data=body, headers=headers, timeout=140)

    if not response.ok:
        print(f"Request unsuccessful: {response.text}, response code: {response.status_code}")
        return None

    my_dict = xmltodict.parse(response.content)


    # Convert from dictionary to dataframe format
    soil_df = None
    try:
        soil_df = pd.DataFrame.from_dict(
            my_dict['soap:Envelope']['soap:Body']['RunQueryResponse']['RunQueryResult']['diffgr:diffgram'][
                'NewDataSet'][
                'Table'])
    except ValueError as e:
        pass
    if isinstance(soil_df, pd.DataFrame):
        if summarytable:
            df = soil_df.drop_duplicates(subset=['componentname'])
            summarytable = df[["componentname", 'prcent', 'chkey']]
            print("summary of the returned soil tables \n")
            print(summarytable)
        # select the dominat componet
        dom_component = soil_df[soil_df.prcent == soil_df.prcent.max()]
        # select by component name
        if select_componentname in soil_df.componentname.unique():
            componentdf = soil_df[soil_df.componentname == select_componentname]
            return componentdf
        elif select_componentname == 'domtcp':
            return dom_component
            # print("the following{0} soil components were found". format(list(soil_df.componentname.unique())))
        elif select_componentname is None:
            # print("the following{0} soil components were found". format(list(soil_df.componentname.unique())))
            return soil_df
        elif select_componentname != 'domtcp' and select_componentname not in soil_df.componentname.unique() or select_componentname != None:
            print(
                f'Ooops! we realised that your component request: {select_componentname} does not exists at the '
                f'specified location. We have returned the dorminant component name')
            return dom_component


# test the function
# aa= DownloadsurgoSoiltables([-90.72704709, 40.93103233],'Osco', summarytable = True)


##Making APSIM soil profile starts here=============================
len_layers = 10
a = 1.35
b = 1.4


# create a variable profile constructor
def soilvar_perdep_cor(nlayers, soil_bottom=200, a=0.5, b=0.5):  # has potential to cythonize
    depthn = np.arange(1, nlayers + 1, 1)
    if a < 0:
        print("Target parameter can not be negative")  # a * e^(-b * x).
    elif (a > 0 and b != 0):
        ep = -b * depthn
        term1 = (a * depthn) * np.exp(ep)
        result = term1 / term1.max()
        return (result)
    elif (a == 0 and b != 0):
        ep = -b * depthn
        result = np.exp(ep) / np.exp(-b)
        return result
    elif (a == 0, b == 0):
        ans = [1] * len_layers
        return ans


def set_depth(depththickness):
    """
  parameters
  depththickness (array):  an array specifying the thicknness for each layer
  nlayers (int); number of layers just to remind you that you have to consider them
  ------
  return
bottom depth and top depth in a turple
  """
    # thickness  = np.tile(thickness, 10
    thickness_array = np.array(depththickness)
    bottomdepth = np.cumsum(thickness_array)  # bottom depth should nothave zero
    top_depth = bottomdepth - thickness_array
    return bottomdepth, top_depth


distparms = {'a': 0, 'b': 0.2}


class OrganiseSoilProfile:
    # Iinitiate the soil object and covert all into numpy array and change them to floating values
    def __init__(self, sdf, thickness, thickness_values=None, bottomdepth=200, state='Iowa'):
        """_summary_

        Args:
            :param sdf (pandas data frame): soil table downloaded from SSURGO_
            :param thickness double: _the thickness of the soil depth e.g 20cm_
            :param bottomdepth (int, optional): _description_. Defaults to 200.
            :param thickness_values: (list or None) optional if provided extrapolation will be based on those vlue and should be the same length as the existing profile depth
         """
        sdf1 = sdf.drop_duplicates(subset=["topdepth"])
        surgodf = sdf1.sort_values('topdepth', ascending=True)
        csr = surgodf.get("CSR").dropna()
        if state == 'Iowa':
            if not csr.empty:
                self.CSR = csr.astype("float")
                
            else:
                self.CSR = None
        else:
            self.CSR = None
        self.clay = npar(surgodf.clay).astype(np.float16)
        self.sand = npar(surgodf.sand).astype(np.float16)
        self.silt = npar(surgodf.silt).astype(np.float16)
        self.OM = npar(surgodf.OM).astype(np.float16)
        self.topdepth = npar(surgodf.topdepth).astype(np.float16)
        self.bottomdepth = npar(surgodf.bottomdepth).astype(np.float16)
        self.BD = npar(surgodf.bb).astype(np.float16)
        self.DUL = npar(surgodf.DUL).astype(np.float64)
        self.L15 = npar(surgodf.L15).astype(np.float64)
        self.PH = npar(surgodf.pH).astype(np.float64)
        self.PAW = npar(surgodf.PAW).astype(np.float16)
        self.saturatedhudraulic_conductivity = npar(surgodf.sat_hidric_cond).astype(np.float16)
        self.KSAT = npar(surgodf.KSAT).astype(np.float16)
        self.K_SAT_r = npar(surgodf.sat_hidric_cond).astype(np.float16)
        self.particledensity = npar(surgodf.pd).astype(np.float16)
        self.muname = surgodf.muname
        self.musymbol = surgodf.musymbol
        self.cokey = surgodf.cokey
        self.slope = surgodf.slope_r
        self.componentname = surgodf.componentname
        self.Nlayers = bottomdepth / thickness
        self.thickness = thickness * 10
        self.wat_r = surgodf.wat_r
        self.chkey = surgodf.chkey
        self.componentpercent = surgodf.prcent
        self.newtopdepth = np.arange(0, bottomdepth, thickness)
        self.newbottomdepth = np.arange(thickness, bottomdepth + thickness, thickness)
        # trial
        self.thickness_values = np.array(thickness_values)

    # create a function that creates a variable profile of the provide _variables
    @staticmethod
    def set_depth(depththickness):
        """
        parameters
        depththickness (array):  an array specifying the thicknness for each layer
        nlayers (int); number of layers just to remind you that you have to consider them
        ------
        return
      bottom depth and top depth in a turple
        """
        # thickness  = np.tile(thickness, 10
        thickness_array = np.array(depththickness)
        bottomdepth = np.cumsum(thickness_array)  # bottom depth should not have zero
        top_depth = bottomdepth - thickness_array
        return bottomdepth, top_depth

    def variable_profile(self, y, kind='linear'):  # has potential to cythonize
        # use the lower depth boundary because divinding by zeros is nearly impossible
        x = self.bottomdepth
        # replace x with the anticipated value of the the new soil depth
        nlayers = self.Nlayers
        xranges = x.max() - x.min()
        newthickness = xranges / nlayers
        # new depth variable for interpolation
        xnew = np.arange(x.min(), x.max(), newthickness)
        # create an interpolation function
        yinterpo = interpolate.interp1d(x, y, kind=kind, assume_sorted=False, fill_value='extrapolate')
        if isinstance(self.thickness_values, str):  # just put a string to evaluate to false i will fix it later
            tv = np.array(self.thickness_values)
            tv = tv.astype('float64')
            xnew, top_dep = OrganiseSoilProfile.set_depth(tv)
        apsimvar = yinterpo(xnew)
        return apsimvar

    def decreasing_exponential_function(self, x, a, b):  # has potential to cythonize
        """
          Compute the decreasing exponential function y = a * e^(-b * x).

          Parameters:
              x (array-like): Input values.
              a (float): Amplitude or scaling factor.
              b (float): Exponential rate.

          Returns:
              numpy.ndarray: The computed decreasing exponential values.
          """
        func = a * np.exp(-b * x)
        return func

    def optimize_exponetial_data(self, x_data, y_data, initial_guess=[0.5, 0.5],
                                 bounds=([0.1, 0.01], [np.inf, np.inf])):  # defaults for carbon

        best_fit_params, _ = curve_fit(self.decreasing_exponential_function, x_data, y_data, p0=initial_guess,
                                       bounds=bounds)
        a_fit, b_fit = best_fit_params
        predicted = self.decreasing_exponential_function(x_data, a_fit, b_fit)
        return predicted

    def interpolated_BD(self):
        if any(elem is None for elem in self.BD):
            return self.variable_profile(self.BD)
        else:
            return np.array(self.BD)

    def cal_satfromBD(self):  # has potential to cythonize
        if any(elem is None for elem in self.particledensity):
            pd = 2.65
            bd = self.interpolated_BD()

            sat = ((2.65 - bd) / 2.65) - 0.02
            sat = self.variable_profile(sat)
            return sat
        else:
            pd = self.particledensity
            bd = self.interpolated_BD()

            sat = ((2.65 - bd) / pd) - 0.02

            sat = self.variable_profile(sat)
            return sat
            # Calculate DUL from sand OM and clay

    def cal_dulFromsand_clay_OM(self):  # has potential to cythonize
        clay = self.clay * 0.01
        sand = self.sand * 0.01
        om = self.OM * 0.01
        ret1 = -0.251 * sand + 0.195 * clay + 0.011 * om + (0.006) * sand * om - 0.027 * clay * om + 0.452 * (
                sand * clay) + 0.299
        dul = ret1 + 1.283 * np.float_power(ret1, 2) - 0.374 * ret1 - 0.015
        dulc = self.variable_profile(dul)
        return dulc

    def cal_l15Fromsand_clay_OM(self):  # has potential to cythonize
        clay = self.clay * 0.01
        sand = self.sand * 0.01
        om = self.OM * 0.01
        ret1 = -0.024 * sand + (
                0.487 * clay) + 0.006 * om + 0.005 * sand * om + 0.013 * clay * om + 0.068 * sand * clay + 0.031
        ret2 = ret1 + 0.14 * ret1 - 0.02
        l151 = self.variable_profile(ret2)
        return l151

    def calculateSATfromwat_r(self):  # has potential to cythonize
        if not all(elem is None for elem in npar(self.particledensity)):
            wat = self.wat_r
            return self.variable_profile(wat)

    def cal_KS(self):  # has potential to cythonize
        ks = self.variable_profile(self.saturatedhudraulic_conductivity)

        ks = ks * (npar([1e-06]) * (60 * 60 * 24) * 1000)

        return ks

    def cal_Carbon(self):  # has potential to cythonize
        ## Brady and Weil (2016)
        carbonn = self.variable_profile(self.OM) / 1.72
        # if carbonans >=3.5:
        #   carbonn = carbonans *0.3
        # else:
        #   carbonn = carbonans
        # print(carbonn)

        xdata = self.newbottomdepth
        try:
            predicted = self.optimize_exponetial_data(xdata, carbonn)

            return predicted
        except Exception as e:
            print("error occured while optimizing soil carbon to the new depth \nplease see:", repr(e))
            return carbonn

    def interpolate_clay(self):
        return self.variable_profile(self.clay)

    def interpolate_silt(self):
        return self.variable_profile(self.silt)

    def interpolate_sand(self):
        return self.variable_profile(self.sand)

    def interpolate_PH(self):
        return self.variable_profile(self.PH)

    def get_L15(self):
        if any(np.isnan(self.L15)):
            L15 = self.cal_l15Fromsand_clay_OM()
            return L15
        else:
            l1 = self.L15 * 0.01
            L15 = self.variable_profile(l1)
            return L15

    def get_DUL(self):
        if any(np.isnan(self.L15)):
            L15i = self.cal_l15Fromsand_clay_OM()

        else:
            l1 = self.L15 * 0.01
            L15i = self.variable_profile(l1)

        if any(np.isnan(self.DUL)):
            DUL = self.cal_dulFromsand_clay_OM()
            return DUL
        else:
            if not any(np.isnan(self.PAW)):
                paw = self.variable_profile(self.PAW)
                DUL = L15i + paw
                return DUL

    def get_AirDry(self):
        if any(np.isnan(self.L15)):
            air = self.cal_l15Fromsand_clay_OM()

            air[0] = air[0] * 0.5
            air[1] = air[1] * 0.9
            return air
        else:
            air = self.L15 / 100
            air[0] = air[0] * 0.5
            air[1] = air[1] * 0.9
            airl = self.variable_profile(air)

            return airl

    def getBD(self):
        return self.variable_profile(self.BD)

    def get_CSR(self):
        if self.CSR:
            return np.array(self.CSR)

    @staticmethod
    def adjust_SAT_BD_DUL(SAT, BD, DUL):
        """
        Adjusts saturation and bulk density values in a NumPy array to meet specific criteria.

        Parameters:
        SAT: 1-D numpy array
        BD: 1-D numpy array
        - target_saturation_a (float): The maximum acceptable saturation value for Soil water Module.
        - target_saturation_b (float): The maximum acceptable saturation value for SWIM
        - target_bulk_density (float): The maximum acceptable bulk density value.

        Returns:
        - np.array: Adjusted 2D NumPy array with saturation and bulk density values.
        """
        # Iterate through each row in the array
        for counter, (sat, dul) in enumerate(zip(SAT, DUL)):
            if dul>=sat:
                diff = dul-sat
                DUL[counter] = sat + diff +0.02
        return SAT, BD, DUL

    def create_soilprofile(self):
        n = int(self.Nlayers)
        Depth = []
        for i in range(len(self.newbottomdepth)):
            Depth.append(str(self.newtopdepth[i]) + "-" + str(self.newbottomdepth[i]))
        Depth = Depth
        Carbon = self.cal_Carbon()
        AirDry = self.get_AirDry()
        L15 = self.get_L15()
        DUL = self.get_DUL()
        if not all(elem is None for elem in npar(self.wat_r)):

            SAT = self.calculateSATfromwat_r() * 0.01

            for i in range(len(SAT)):
                if SAT[i] < DUL[i]:
                    SAT[i] = DUL[i] + 0.02
            else:
                SAT[i] = SAT[i]
            BD = (1 - SAT) * 2.65
            for i in range(len(SAT)):  # added it
                if SAT[i] > 0.381 and BD[i] >= 1.639:
                    SAT[i] = 0.381 - 0.001
                    # print(SAT[i])
        else:
            SAT = self.cal_satfromBD()
            BD = self.getBD()
            for i in range(len(SAT)):
                if SAT[i] < DUL[i]:
                    SAT[i] = DUL[i] + 0.001
            else:
                SAT[i] = SAT[i]
            for i in range(len(SAT)):  # added it
                if SAT[i] > 0.381 and BD[i] >= 1.639:
                    SAT[i] = 0.381 - 0.01
                    # print(SAT[i])
        # adjust layer 9 issues associated with SAT and BD
        SAT, BD, DUL = self.adjust_SAT_BD_DUL(SAT, BD, DUL)
        KS = self.cal_KS()
        PH = self.interpolate_PH()
        ParticleSizeClay = self.interpolate_clay()
        ParticleSizeSilt = self.interpolate_silt()
        ParticleSizeSand = self.interpolate_sand()
        VA = [SAT, BD, DUL, KS, PH, ParticleSizeClay, ParticleSizeSilt, ParticleSizeSand, L15, Depth, Carbon, PH]
        DUL = np.sort(DUL)[::-1]
        SAT = np.sort(SAT)[::-1]

        def calculate_bd(sat):
            bd = 2.65 - (sat + 0.02) * 2.65
            return bd

        BD = calculate_bd(SAT)

        df = pd.DataFrame(
            {"Depth": Depth, "Thickness": [self.thickness] * n, "BD": BD, "AirDry": AirDry, "LL15": L15, "DUL": DUL,
             "SAT": SAT, "KS": KS, "Carbon": Carbon,
             "PH": PH, "ParticleSizeClay": ParticleSizeClay, "ParticleSizeSilt": ParticleSizeSilt,
             "ParticleSizeSand": ParticleSizeSand})

        return df

    def exponential_function_inc_yvalue(x, a, b):
        return a * np.exp(-b * x)

    def optimize_exp_increasing_y_values(self, y_data):
        if isinstance(self.thickness_values, np.ndarray):

            xdata = np.cumsum(np.array(self.thickness_values))
        else:
            xdata = np.cumsum(np.array(self.newbottomdepth))
        xdata = xdata.astype('float64')

        # Sort the data points by increasing y-values
        sorted_indices = np.argsort(y_data)
        sorted_x_data = xdata[sorted_indices]
        sorted_y_data = y_data[sorted_indices]
        # Perform curve fitting with the sorted data
        best_fit_params, _ = curve_fit(self.exponential_function_inc_yvalue, sorted_x_data, sorted_y_data)

        # Extract the best-fit parameters (a_fit and b_fit)
        a_fit, b_fit = best_fit_params
        predcited = self.exponential_function_inc_yvalue(xdata, a_fit, b_fit)

        # Plot the original data and the fitted curve
        plt.scatter(xdata, y_data, label='Original Data')
        plt.plot(xdata, self.exponential_function_inc_yvalue(xdata, a_fit, b_fit), label='Fitted Curve', color='red')
        plt.xlabel('x')
        plt.ylabel('y')
        plt.legend()
        plt.show()
        return predcited

    def cal_missingFromSurgo(self, curveparam_a=0, curveparam_b=0.2, crops=["Wheat", "Maize", "Soybean", "Rye"],
                             metadata=None, soilwat=None, swim=None, soilorganicmatter=None):
        nlayers = int(self.Nlayers)
        # ad  = self.get_AirDry()[1]
        cropLL = self.get_AirDry()

        # Original thought
        # ad * soilvar_perdep_cor(nlayers, a = curveparam_a, b = curveparam_b)
        cropKL = 0.08 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=curveparam_b)

        cropXF = 1 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=0)

        # create a data frame for these three _variables
        dfs = pd.DataFrame({'kl': cropKL, 'll': cropLL, 'xf': cropXF})
        SoilCNRatio = np.full(shape=nlayers, fill_value=12, dtype=np.int64)
        FOM = 150 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=curveparam_b)
        FOMCN = np.full(shape=nlayers, fill_value=40, dtype=np.int64)
        FBiom = 0.045 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=curveparam_b)
        Fi = 0.83 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=-0.01)
        # to do later
        # try:
        #   predicted= self.optimize_exp_increasing_y_values(np.array(Fi))
        #   FInert = predicted
        # except Exception as e:
        #   print("error optiming FInert value encountered;", repr(e))
        FInert = Fi

        NO3N = 0.5 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=0.01)
        NH4N = 0.05 * soilvar_perdep_cor(nlayers, a=curveparam_a, b=0.01)
        FInert[0] = 0.65
        FInert[1] = 0.668
        FBiom[0] = 0.0395
        FBiom[1] = 0.035
        # from above
        Carbon = self.cal_Carbon()
        PH = self.interpolate_PH()

        # vn = npar
        # PH = 6.5 * soilvar_perdep_cor(nlayers, a = curveparam_a, b = 0)

        organic = pd.DataFrame(
            {'Carbon': Carbon, 'SoilCNRatio': SoilCNRatio, 'cropLL': cropLL, 'cropKL': cropKL, 'FOM': FOM,
             'FOM.CN': FOMCN, 'FBiom': FBiom, 'FInert': FInert, 'NO3N': NO3N, 'NH4N': NH4N, 'PH': PH})
        # create a list to store the crop names
        names = []
        for i in crops:
            names.append([i + "KK", i + 'LL ', i + 'XF'])
        cropframe = []
        for i in names:
            cropframe.append(dfs.rename(columns={"kl": i[0], "ll": i[1], "xf": i[2]}))
        # print(cropframe)
        # pd.concat is faster
        cropdf = pd.concat(cropframe, join='outer', axis=1)
        physical = self.create_soilprofile()

        po = 1 - (physical["BD"] / 2.65)
        swi_con = (po - physical['DUL']) / po

        # create a alist
        frame = [physical, organic, cropdf, metadata, swi_con, swi_con]
        # All soil data frames
        if isinstance(self.CSR, (list, np.ndarray, pd.Series)):
            CSR = np.tile(self.CSR.iloc[0], int(self.Nlayers))
        else:
            CSR = None
        resultdf = pd.concat(frame, join='outer', axis=1)
        finalsp = {'soil ': resultdf, 'crops': crops, 'metadata': metadata, 'soilwat': soilwat, 'swim': swim,
                   'soilorganicmatter ': soilorganicmatter, 'SWCON': swi_con}
        # return pd.DataFrame(finalsp)
        frame.insert(3, CSR, )

        return frame

OrganizeAPSIMsoil_profile = OrganiseSoilProfile #for backward compatibiltiy
if __name__ == '__main__':
    lonlat = -92.097702, 41.8780025
    dw = DownloadsurgoSoiltables(lonlat)
    sop = OrganiseSoilProfile(dw, 20)
    data = sop.cal_missingFromSurgo()
    from apsimNGpy.core.apsim import ApsimModel
    from apsimNGpy.core.base_data import load_default_simulations
    from pathlib import Path
    from apsimNGpy import settings

    ap_sim = load_default_simulations(simulations_object=False)

    m = ApsimModel(ap_sim, thickness_values=settings.SOIL_THICKNESS)
    m.replace_downloaded_soils(data, m.extract_simulation_name)

    m.run("MaizeR")




import os

from apsimNGpy.manager import weathermanager

from apsimNGpy.core.base_data import load_default_simulations
from apsimNGpy.core.apsim import ApsimModel as SoilModel
from pathlib import Path
import os

# set the working path
cwd = Path.cwd().home()
os.chdir(cwd)
CROP = "Maize"
# download weather from nasa
# sample lonlat
lonlat = -3.01183333, 42.08333333
weather = weathermanager.get_weather(lonlat, start=1990, end=2020, filename="somewhere.met", source='nasa')
data = load_default_simulations(crop=CROP, simulations_object=True)

# initialise
apsim = SoilModel(data.path)
# Now insert weather file into the model
apsim.replace_met_file(weather_file=weather)
# check if it was successful
print(apsim.show_met_file_in_simulation())
# try running
apsim.run()
# collect the results
res = apsim.results
print(res)
# save your model
apsim.save(file_name='changed_met.apsimx')
# check the saved path
print(apsim.path)
# in other cases, you don't need to collect results to show that things are. you can check through ran_ok property
print(apsim.ran_ok)



"""
Interface to APSIM simulation models using Python.NET
author: Richard Magala
email: magalarich20@gmail.com
"""
import logging, pathlib
from typing import Union
import os
import numpy as np
import time
from apsimNGpy.manager.soilmanager import DownloadsurgoSoiltables, OrganiseSoilProfile
import apsimNGpy.manager.weathermanager as weather
import pandas as pd
import sys
# prepare for the C# import
# from apsimNGpy.core.pythonet_config import start_pythonnet

from apsimNGpy.core.core import CoreModel, Models
from apsimNGpy.core.inspector import Inspector
# now we can safely import any c# related libraries
from System.Collections.Generic import *
from Models.Core import Simulations
from System import *
from Models.Soils import Solute, Water, Chemical
from Models.Soils import Soil, Physical, SoilCrop, Organic, LayerStructure

from typing import Union

# constants
REPORT_PATH = {'Carbon': '[Soil].Nutrient.TotalC/1000 as dyn', 'DUL': '[Soil].SoilWater.PAW as paw', 'N03':
    '[Soil].Nutrient.NO3.ppm as N03'}


# decorator to monitor performance
def timing_decorator(func):
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        elapsed_time = end_time - start_time
        print(f"{func.__name__} took {elapsed_time:.4f} seconds to execute.")
        return result

    return wrapper


class ApsimModel(CoreModel):
    """
    Main class for apsimNGpy modules.
    It inherits from the CoreModel class and therefore has access to a repertoire of methods from it.

    This implies that you can still run the model and modify parameters as needed.
    Example:
        >>> from apsimNGpy.core.apsim import ApsimModel
        >>> from apsimNGpy.core.base_data import load_default_simulations
        >>> path_model = load_default_simulations(crop='Maize', simulations_object=False)
        >>> model = ApsimModel(path_model, set_wd=Path.home())# replace with your path
        >>> model.run(report_name='Report') # report is the default replace as needed
    """
    def __init__(self, model: Union[os.PathLike, dict, str], out_path: os.PathLike = None, out: os.PathLike = None,
                 lonlat: tuple = None, soil_series: str = 'domtcp', thickness: int = 20, bottomdepth: int = 200,

                 thickness_values: list = None, run_all_soils: bool = False, set_wd=None, **kwargs):
        super().__init__(model, out_path, set_wd, **kwargs)
        self.soiltype = None
        self.SWICON = None
        """get SSURG soil tables and organise it to aPSim soil profiles -------------------- parameters soil_series (
        _str_) specifying the soils series default is domcp meaning dominant soil series, users can specify any 
        other series if known. lonlat {_lonlat_} of the specified location run_all_soils (__boolean_) if True, 
        all soil series in the map units will be analysed if false the soil series will be the one determined by 
        parameter soil series model is a path of aPSim file or aPSIMNG object"""
        self.lonlat = lonlat
        self.Nlayers = bottomdepth / thickness
        bm = bottomdepth * 10
        if thickness_values is None:
            thickness_values = self.auto_gen_thickness_layers(max_depth=bm, n_layers=int(self.Nlayers))
        self.soil_series = soil_series
        self.thickness = thickness
        self.out_path = out_path or out
        self._thickness_values = thickness_values
        self.copy = True
        self.run_all_soils = run_all_soils
        if not isinstance(thickness_values, np.ndarray):
            self.thickness_values = np.array(self._thickness_values, dtype=np.float64)  # apsim uses floating digit number
        else:
            self.thickness_values = self._thickness_values
        if kwargs.get('experiment', False):
            self.create_experiment()
    @property
    def thickness_values(self):
        return self._thickness_values
    @thickness_values.setter
    def thickness_values(self, values):
        self._thickness_values = values

    def show_cropsoil_names(self, simulations):
        for simu in self.find_simulations(simulations):
            pysoil = simu.FindDescendant[Physical]()
            soil_crop = pysoil.FindAllDescendants[SoilCrop]()
            # can be use to target specific crop
            for cropLL in soil_crop:
                print(cropLL.Name)

    def _replace_cropsoil_names(self, simulations, existing_crop_names, new_cropname):
        for simu in self.find_simulations(simulations):
            pysoil = simu.FindDescendant[Physical]()
            soil_crop = pysoil.FindAllDescendants[SoilCrop]()

            for crops in soil_crop:
                if crops.Name == existing_crop_names:
                    crops.Name = new_cropname



    def adjust_dul(self, simulations: Union[tuple, list] = None):
        """
        - This method checks whether the soil ``SAT`` is above or below ``DUL`` and decreases ``DUL``  values accordingly
        - Need to call this method everytime ``SAT`` is changed, or ``DUL`` is changed accordingly.

        ``simulations``: str, name of the simulation where we want to adjust DUL and SAT according.

        ``returns``:
            model object
        """
        duL = self.extract_any_soil_physical('DUL', simulations)
        print(duL)
        saT = self.extract_any_soil_physical('SAT', simulations)
        for sim in duL:
            duls = duL[sim]

            sats = saT[sim]
            for enum, (s, d) in enumerate(zip(sats, duls)):
                # first check if they are equal
                if d >= s:
                    # if d is greater than s, then by what value, we need this value to add it to 0.02
                    #  to be certain all the time that dul is less than s we subtract the summed value
                    diff = d - s
                    duls[enum] = d - (diff + 0.02)
                    print(d)
                else:
                    duls[enum] = d
            if not simulations:
                soil_object = self.Simulations.FindDescendant[Soil]()

                soilsy = soil_object.FindDescendant[Physical]()
                soilsy.DUL =duls
        self.save()
            #self.replace_any_soil_physical('DUL', simulations, duL)
        return self

    def _get_SSURGO_soil_profile(self, lonlat: tuple, run_all_soils: bool = False):
        self.lonlat = None
        self.lonlat = lonlat
        self.dict_of_soils_tables = {}
        if not self.run_all_soils:
            self.soil_tables = DownloadsurgoSoiltables(self.lonlat, select_componentname=self.soil_series)
            for ss in self.soil_tables.componentname.unique():
                self.dict_of_soils_tables[ss] = self.soil_tables[self.soil_tables['componentname'] == ss]

        if self.run_all_soils:
            self.dict_of_soils_tables = {}
            self.soil_tables = DownloadsurgoSoiltables(self.lonlat)

            self.percent = self.soil_tables.prcent.unique()
            # create a dictionary of soil series
            self.unique_soil_series = self.soil_tables.componentname.unique()

            new_col = []
            for i in range(len(self.soil_tables['chkey'])):
                xi = str(list(self.soil_tables['chkey'])[i]) + "-" + list(self.soil_tables['componentname'])[i]
                new_col.append(xi)
            self.soil_tables["ch_comp"] = list(new_col)
            self.grouped = self.soil_tables.groupby('ch_comp')['prcent'].unique().apply(lambda x: x[0])
            self.component_percent_dict = self.grouped.to_dict()
            for ss in self.soil_tables.componentname.unique():
                self.dict_of_soils_tables[ss] = self.soil_tables[self.soil_tables['componentname'] == ss]
        return self

    @staticmethod
    def get_weather_online(lonlat: tuple, start: int, end: int):
        wp = weather.get_met_from_day_met(lonlat, start=start, end=end)
        wpath = os.path.join(os.getcwd(), wp)
        return wpath

    @property
    def get_unique_soil_series(self):
        """this function collects the unique soil types

        Args:
            lonlat (_tuple_): longitude and latitude of the target location

        """
        try:
            soil_tables = DownloadsurgoSoiltables(self.lonlat)
            pr = soil_tables.prcent.unique()

            grouped = soil_tables.groupby('componentname')[
                'prcent'].unique()  # .agg(list) #.unique().apply(lambda x: x[0])
            component_percent_dict = grouped.to_dict()
            return component_percent_dict
        except Exception as e:
            raise

    def replace_soils(self, lonlat: tuple, simulation_names: Union[tuple, list], verbose=False):
        self.thickness_replace = None
        if isinstance(self.thickness_values, np.ndarray):  # since it is alreaduy converted to an array
            self.thickness_replace = self.thickness_values

        else:
            tv = np.tile(float(self.thickness), int(self.Nlayers))
            self.thickness_replace = tv
        pss = self._get_SSURGO_soil_profile(lonlat)
        for keys in pss.dict_of_soils_tables.keys():
            self.soiltype = keys
            if verbose:
                print("Padding variabales for:", keys)
            self.soil_profile = OrganiseSoilProfile(self.dict_of_soils_tables[keys],
                                                    thickness_values=self.thickness_values,
                                                    thickness=self.thickness)
            missing_properties = self.soil_profile.cal_missingFromSurgo()  # returns a list of physical, organic and cropdf each in a data frame
            physical_calculated = missing_properties[0]
            self.organic_calcualted = missing_properties[1]
            self.cropdf = missing_properties[2]
            # ps = self._get_SSURGO_soil_profile()

            # self.thickness_replace = list(np.full(shape=int(self.Nlayers,), fill_value=self.thickness*10,  dtype=np.float64))
            for simu in self.find_simulations(simulation_names):
                pysoil = simu.FindDescendant[Physical]()  # meaning physical soil child
                soil_crop = pysoil.FindChild[SoilCrop]()
                water = simu.FindDescendant[Water]()  # for the crop water parameters
                soil_crop.LL = physical_calculated.AirDry
                pysoil.DUL = physical_calculated.DUL
                pysoil.SAT = physical_calculated.SAT
                pysoil.BD = physical_calculated.BD
                pysoil.KS = physical_calculated.KS
                pysoil.LL15 = physical_calculated.LL15
                pysoil.ParticleSizeClay = physical_calculated.ParticleSizeClay
                pysoil.ParticleSizeSand = physical_calculated.ParticleSizeSand
                pysoil.ParticleSizeSilt = physical_calculated.ParticleSizeSilt
                water.InitialValues = physical_calculated.DUL

                water.Thickness = self.thickness_replace
                pysoil.AirDry = soil_crop.LL
                pysoil.Thickness = self.thickness_replace
                # replace the organic soils
            for simu in self.find_simulations(simulation_names):
                organic = simu.FindDescendant[Organic]()
                organic.Thickness = self.thickness_replace
                organic.SoilCNRatio = self.organic_calcualted.SoilCNRatio
                organic.FBiom = self.organic_calcualted.FBiom
                organic.FOM = self.organic_calcualted.FOM
                organic.FInert = self.organic_calcualted.FInert
                organic.Carbon = self.organic_calcualted.Carbon
                chemical = simu.FindDescendant[Chemical]()
                chemical.Thickness = self.thickness_replace
                # to do fix the crop management may use FindAllDescendants. it worked
                # XF = np.full(shape=self.Nlayers, fill_value=1,  dtype=np.float64)
                XF = np.tile(float(1), int(self.Nlayers))

            for simu in self.find_simulations(simulation_names):
                soil_crop = pysoil.FindAllDescendants[SoilCrop]()
                # can be use to target specific crop
                for cropLL in soil_crop:
                    cropLL.LL = pysoil.AirDry
                    cropLL.KL = self.organic_calcualted.cropKL
                    cropLL.XF = XF
                    cropLL.Thickness = self.thickness_replace
            if verbose:
                print("soil replacement complete")
            # self.run()
        return self
        # print(self.results)

    def replace_downloaded_soils(self, soil_tables: Union[dict, list], simulation_names: Union[tuple, list], **kwargs):
        """
            Updates soil parameters and configurations for downloaded soil data in simulation models.

            This method adjusts soil physical and organic parameters based on provided soil tables and applies these
            adjustments to specified simulation models.

            Parameters:
            ``soil_tables`` (list): A list containing soil data tables. Expected to contain: see the naming
            convention in the for APSIM - [0]: DataFrame with physical soil parameters. - [1]: DataFrame with organic
            soil parameters. - [2]: DataFrame with crop-specific soil parameters. - simulation_names (list of str): Names or identifiers for the simulations to
            be updated.s


            Returns:
            - self: Returns an instance of the class for ``chaining`` methods.

            This method directly modifies the simulation instances found by ``find_simulations`` method calls,
            updating physical and organic soil properties, as well as crop-specific parameters like lower limit (``LL``),
            drain upper limit (``DUL``), saturation (``SAT``), bulk density (``BD``), hydraulic conductivity at saturation (``KS``),
            and more based on the provided soil tables.

    ->> key-word argument

            ``set_sw_con``: Boolean, set the drainage coefficient for each layer
            ``adJust_kl``:: Bollean, adjust, kl based on productivity index
            ``CultvarName``: cultivar name which is in the sowing module for adjusting the rue
            ``tillage``: specify whether you will be carried to adjust some physical parameters

        """

        self.thickness_replace = self.thickness_values
        physical_calculated = soil_tables[0]
        self.organic_calcualted = soil_tables[1]
        self.cropdf = soil_tables[2]
        self.SWICON = soil_tables[6]  # TODO To put these tables in the dictionary isn't soilmanager module
        for simu in self.find_simulations(simulation_names):
            pysoil = simu.FindDescendant[Physical]()  # meaning physical soil child

            soil_crop = pysoil.FindChild[SoilCrop]()
            water = simu.FindDescendant[Water]()  # for the crop water parameters
            soil_crop.LL = physical_calculated.AirDry
            pysoil.DUL = physical_calculated.DUL
            pysoil.SAT = physical_calculated.SAT
            pysoil.BD = physical_calculated.BD
            pysoil.KS = physical_calculated.KS
            pysoil.LL15 = physical_calculated.LL15
            pysoil.ParticleSizeClay = physical_calculated.ParticleSizeClay
            pysoil.ParticleSizeSand = physical_calculated.ParticleSizeSand
            pysoil.ParticleSizeSilt = physical_calculated.ParticleSizeSilt
            water.InitialValues = physical_calculated.DUL

            water.Thickness = self.thickness_replace
            pysoil.AirDry = soil_crop.LL
            pysoil.Thickness = self.thickness_replace
            # print(len(pysoil.Thickness))
            # replace the organic soils
        for simu in self.find_simulations(simulation_names):
            organic = simu.FindDescendant[Organic]()
            organic.Thickness = self.thickness_replace
            organic.SoilCNRatio = self.organic_calcualted.SoilCNRatio
            organic.FBiom = self.organic_calcualted.FBiom
            # you could adjust practices here, but still looking for a cool way to do it
            if kwargs.get('No_till', False):
                self.organic_calcualted.loc[:1, 'FBiom'] = self.organic_calcualted.loc[:1,
                                                           'FBiom'] + 0.2 * self.organic_calcualted.loc[:1, 'FBiom']
                organic.FBiom = self.organic_calcualted.FBiom[:2]
            organic.FOM = self.organic_calcualted.FOM
            organic.FInert = self.organic_calcualted.FInert
            organic.Carbon = self.organic_calcualted.Carbon
            chemical = simu.FindDescendant[Chemical]()
            chemical.Thickness = self.thickness_values
            XF = np.tile(float(1), int(self.Nlayers))

        for simu in self.find_simulations(simulation_names):
            soil_crop = pysoil.FindAllDescendants[SoilCrop]()
            # can be used to target specific crop
            for cropLL in soil_crop:
                cropLL.LL = pysoil.LL15
                kl = self.organic_calcualted.cropKL
                cropLL.KL = kl
                # if kwargs.get('No_till', False):
                #     cropLL.KL = cropLL.KL + np.array([0.2]) * cropLL.KL
                cropLL.XF = XF
                cropLL.Thickness = self.thickness_replace
        for simu in self.find_simulations(simulation_names):
            zone = simu.FindChild[Models.Core.Zone]()
            try:
                swim = pysoil.FindAllDescendants[LayerStructure]()
                swim.LayerStructure = self.thickness_values
            except Exception as e:
                pass
        # replace drainage coefficient for each layer based on DUL and BD
        for sim in self.find_simulations(simulation_names):
            try:
                wb = sim.FindDescendant[Models.WaterModel.WaterBalance]()
                wb.SWCON = self.SWICON
                wb.Thickness = self.thickness_values
                if kwargs.get('CN2Bare', None):
                    wb.CN2Bare = kwargs.get('CN2Bare')
                if kwargs.get('CNRed', None):
                    wb.CNRed = kwargs.get('CNRed')
            except:
                # in the case of sim model, pass

                pass

        return self

    # print(self.results)

    def _change_met_file(self, lonlatmet: tuple = None,
                         simulation_names: Union[tuple, list] = None):  # to be accessed only in this class
        """_similar to class weather management but just in case we want to change the weather within the subclass
        # uses existing start and end years to download the weather data

        """
        if lonlatmet == None:
            self.lonlat = self.lonlat
        else:
            self.lonlat = lonlatmet

        start, end = self.extract_start_end_years()
        wp = weather.get_met_from_day_met(self.lonlat, start, end)
        wpath = os.path.join(os.getcwd(), wp)
        wpath = os.path.join(os.getcwd(), wp)
        if self.simulation_names:
            sim_name = list(self.simulation_names)
        else:
            sim_name = self.simulation_names  # because it is a property decorator
        self.replace_met_file(wpath, sim_name)
        return self

    def run_edited_file(self, table_name=None):
        # to be deprecated
        """

            :param table_name (str): repot table name in the database

        """
        return self.run(report_name=table_name).results

    def spin_up(self, report_name: str = 'Report', start=None, end=None, spin_var="Carbon", simulations=None):
        """
        Perform a spin-up operation on the aPSim model.

        This method is used to simulate a spin-up operation in an aPSim model. During a spin-up, various soil properties or
        _variables may be adjusted based on the simulation results.

        Parameters:
        ----------
        ``report_name`` : str, optional (default: 'Report')
            The name of the aPSim report to be used for simulation results.

        ``start`` : str, optional
            The start date for the simulation (e.g., '01-01-2023'). If provided, it will change the simulation start date.

        ``end`` : str, optional
            The end date for the simulation (e.g., '3-12-2023'). If provided, it will change the simulation end date.

        ``spin_var`` : str, optional (default: 'Carbon'). the difference between the start and end date will determine the spin-up period
            The variable representing the child of spin-up operation. Supported values are 'Carbon' or 'DUL'.

        ``Returns:``
        -------
        self : ApsimModel
            The modified ``ApsimModel`` object after the spin-up operation.
            you could call ``save_edited`` file and save it to your specified location, but you can also proceed with the simulation

        """

        insert_var = REPORT_PATH.get(spin_var)
        if start and end:
            self.change_simulation_dates(start, end)
        for simu in self.find_simulations(simulations):
            pysoil = simu.FindDescendant[Physical]()
        THICKNESS = self.extract_any_soil_physical("Thickness")
        th = np.array(THICKNESS)
        self.change_report(insert_var, report_name=report_name)
        rpn = insert_var.split(" ")[-1]
        self.run(report_name=report_name)
        DF = self.results

        df_sel = DF.filter(regex=r'^{0}'.format(rpn), axis=1)
        df_sel = df_sel.mean(numeric_only=True)
        print(df_sel)
        if spin_var == 'Carbon':
            if 'TotalC' not in insert_var:
                raise ValueError("wrong report variable path: '{0}' supplied according to requested spin up " \
                                 "var".format(insert_var))

            bd = list(pysoil.DUL)

            bd = np.array(bd)
            cf = np.array(bd) * np.array(th)
            cf = np.divide(cf, 1)  # this convert to percentage
            per = np.array(df_sel) / cf
            new_carbon = [i for i in np.array(per).flatten()]

            self.replace_any_soil_organic(spin_var, new_carbon)
        if spin_var == 'DUL':
            if 'PAW' not in insert_var:
                raise ValueError("wrong report variable path: '{0}' supplied according to requested spin up var" \
                                 .format(insert_var))
            l_15 = pysoil.LL15
            ll = np.array(l_15)
            dul = ll + df_sel
            dul = list(np.array(dul).flatten())

            pysoil.DUL = dul

            return self

    def replace_met_from_web(self, lonlat, start_year, end_year, file_name=None):
        if not file_name:
            file_name = self.path.strip(".apsimx") + "_w_.met"
        w_f = weather.get_met_from_day_met(lonlat, start=start_year, end=end_year, filename=file_name)
        wf = os.path.abspath(w_f)
        self.replace_met_file(weather_file=wf)
        return self

    def auto_gen_thickness_layers(self,
            max_depth, n_layers=10, thin_layers=3, thin_thickness=100,
            growth_type="linear", thick_growth_rate=1.5
    ):
        """
        Generate layer thicknesses from surface to depth, starting with thin layers and increasing thickness.

        Args:
            ``max_depth`` (float): Total depth in mm.

            ``n_layers`` (int): Total number of layers.

            ``thin_layers`` (int): Number of initial thin layers.

            ``thin_thickness`` (float): Thickness of each thin layer.

            ``growth_type`` (str): 'linear' or 'exponential'.

            ``thick_growth_rate`` (float): Growth factor for thick layers (e.g., +50% each layer if exponential).

        ``Returns:``
            List[float]: List of layer thicknesses summing to max_depth.
        """
        from math import pow

        assert 0 < thin_layers < n_layers, "thin_layers must be less than total layers"
        assert thin_thickness > 0, "thin_thickness must be positive"
        assert growth_type in ['linear', 'exponential'], "Invalid growth_type"

        thick_layers = n_layers - thin_layers
        thin_total = thin_layers * thin_thickness
        remaining_depth = max_depth - thin_total

        if remaining_depth <= 0:
            raise ValueError("Thin layers consume more than total depth.")

        # --- Step 1: Thin layers ---
        thin_parts = [thin_thickness] * thin_layers

        # --- Step 2: Thick layers ---
        if growth_type == "linear":
            # We'll solve an arithmetic series: a + (a + d) + ... = remaining_depth
            # Assume a = base_thick, d = constant increment
            # So: total = thick_layers * a + d * (0 + 1 + ... + thick_layers-1)
            increments = list(range(thick_layers))
            sum_increments = sum(increments)
            base_thick = (remaining_depth - sum_increments * thick_growth_rate) / thick_layers
            thick_parts = [base_thick + i * thick_growth_rate for i in increments]

        elif growth_type == "exponential":
            # Geometric series: a*r^0 + a*r^1 + ... = remaining_depth
            r = thick_growth_rate
            denom = sum([pow(r, i) for i in range(thick_layers)])
            a = remaining_depth / denom
            thick_parts = [a * pow(r, i) for i in range(thick_layers)]

        # Final result
        layers = thin_parts + thick_parts

        # Precision check
        total = sum(layers)
        if abs(total - max_depth) > 1e-6:
            # Final adjustment to fix floating point error
            layers[-1] += max_depth - total
        layers.sort()
        layers = [int(i) for i in layers]

        return layers



    def replace_soil_profile_from_web(self, **kwargs):
        from apsimNGpy.manager.weathermanager import get_weather, _is_within_USA_mainland
        lon_lat = kwargs.get('lonlat', self.lonlat)
        if not _is_within_USA_mainland(lon_lat):
            raise ValueError(f"{lon_lat} is not within USA. coordnates outside USA are not supported yet")
        thickness = kwargs.get('thickness', 20)
        sim_name = kwargs.get('sim_name', self.simulation_names)
        assert lon_lat, 'Please supply the lonlat'
        sp = DownloadsurgoSoiltables(lon_lat)
        sop = OrganiseSoilProfile(sp, thickness, thickness_values=self.thickness_values).cal_missingFromSurgo()

        if self.thickness_values is None:
            self.thickness_values = self.auto_gen_thickness_layers(max_depth=2200, n_layers=int(self.Nlayers),
                                                    thin_layers=3, thin_thickness=100,
                                                    growth_type='linear', thick_growth_rate=50)
        self.thickness_replace =self.thickness_values.copy()
        self.replace_downloaded_soils(sop, simulation_names=sim_name)
        self.save()
        return self


if __name__ == '__main__':
    # test
    from pathlib import Path

    # Model = FileFormat.ReadFromFile[Models.Core.Simulations](model, None, False)
    os.chdir(Path.home())
    from apsimNGpy.core.base_data import load_default_simulations

    try:
        lonlat = -93.7738, 42.0204
        al =load_default_simulations(simulations_object=False)
        model = al

        from apsimNGpy import settings

        model = ApsimModel(model, out_path=None,
                           thickness_values=settings.SOIL_THICKNESS)
        model.replace_met_from_web(lonlat=lonlat, start_year=2001, end_year=2020)
        from apsimNGpy.manager import soilmanager as sm

        st = sm.DownloadsurgoSoiltables(lonlat)
        sp = sm.OrganiseSoilProfile(st, 20)
        sop = sp.cal_missingFromSurgo()
        model.replace_downloaded_soils(sop, model.simulation_names, No_till=True)
        bd = model.extract_any_soil_physical("BD")

    except Exception as e:
        print(type(e).__name__, repr(e))
        exc_type, exc_value, exc_traceback = sys.exc_info()
        # Extract the line number from the traceback object
        line_number = exc_traceback.tb_lineno
        print(f"Error: {type(e).__name__} occurred on line: {line_number} execution value: {exc_value}")







from pathlib import Path
from os.path import abspath
import os

path = Path("../../")
ax = path.rglob('*.apsimx')
db = path.rglob('*.db')
remove = list(ax) + list(db) + list(path.rglob('*.db-shm')) + list(path.rglob('*.csv'))
for i in remove:
    try:
        os.remove(str(i))
    except PermissionError as e:
        print(f"skipping '{i}' due to permission error")



from apsimNGpy.core.core import Models
model_lists = []
from apsimNGpy.core.base_data import load_default_simulations
model = load_default_simulations(crop = 'maize')


import types

def extract_models(model_namespace=None, visited=None):
    """
    Recursively extracts all model class names from a given namespace,
    diving into all attributes including nested objects, modules, and classes.
    """
    if model_namespace is None:
        model_namespace = Models

    if visited is None:
        visited = set()

    models_found = []

    # Avoid processing the same object multiple times (handle circular refs)
    if id(model_namespace) in visited:
        return []

    visited.add(id(model_namespace))

    # Try to iterate through all attributes of the object
    try:
        attr_names = dir(model_namespace)
    except Exception:
        return []

    for attr_name in attr_names:
        if attr_name.startswith("__"):
            continue

        try:
            attr_value = getattr(model_namespace, attr_name)
        except Exception:
            continue  # Safely skip inaccessible attributes

        # If it's a class, store its full path
        if isinstance(attr_value, type):
            model_path = f"{attr_value.__module__}.{attr_value.__name__}"
            models_found.append(model_path)

        # If it's a module, class, or object with attributes — go deeper
        elif isinstance(attr_value, (types.ModuleType, object)) and not isinstance(attr_value, (int, float, str, list, dict, set, tuple)):
            models_found.extend(extract_models(attr_value, visited))

    return sorted(set(models_found))


if __name__ == "__main__":
    print(extract_models())
    from collections import defaultdict
    model.add_crop_replacements("Maize")
    model.create_experiment()
    # Your flat list of paths
    paths = model.inspect_model(Models.Core.Model)
    def filter_out():
        import Models
        data = []
        nd = ['Models.Core.Simulation', 'Models.Soils.Soil', 'Models.PMF.Plant', 'Models.Manager',
              'Models.Climate.Weather', 'Models.Report', 'Models.Clock', 'Models.Core.Folder', 'Models.Soils.Solute',
              'Models.Soils.Swim3','Models.Soils.SoilCrop', 'Models.Soils.Water','Models.Summary', 'Models.Core.Zone',
              'Models.Soils.CERESSoilTemperature','Models.Series','Models.Factorial.Experiment', 'Models.Factorial.Permutation',
              'Models.Factorial.Factors',
              'Models.Sobol','Models.Operations', 'Models.Morris', 'Models.Fertiliser','Models.Core.Events','Models.Core.VariableComposite',
              'Models.Soils.Physical', 'Models.Soils.Chemical', 'Models.Soils.Organic']
        for i in nd:

            ans =model.inspect_model(eval(i))
            if not 'Replacements' in ans and 'Folder' in i:
                continue
            data.extend(ans)
        return data
    paths = filter_out()
    # Build a nested tree structure
    def build_tree(paths):
        tree = lambda: defaultdict(tree)
        root = tree()
        for path in paths:
            parts = path.strip('.').split('.')
            current = root
            for part in parts:
                current = current[part]
        return root


    # Recursively print the tree
    def print_tree(node, indent=0, full_path=""):
        for key in sorted(node.keys()):
            current_path = f"{full_path}.{key}" if full_path else key
            print("    " * indent + f"- {key}: .{current_path}")
            print_tree(node[key], indent + 1, current_path)


    # Build and print
    tree = build_tree(paths)
    print_tree(tree)



import logging
from collections import deque

import numpy as np
import glob
import os
from os.path import join as opj
import shutil
import random
import string
import geopandas as gpd
from shapely.geometry import Polygon
from shapely.ops import unary_union
from shapely.geometry import Point
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from threading import Thread
from scipy.optimize import curve_fit
from time import perf_counter
import pandas as pd
import time
from functools import cache
from pathlib import Path
from shapely import wkt
from geopy.distance import geodesic
import functools
import traceback
import sys
from apsimNGpy.settings import logger
from platform import system
from subprocess import call, run, Popen
def select_process(use_thread, ncores):
    return ThreadPoolExecutor(ncores) if use_thread else ProcessPoolExecutor(ncores)

def open_file_in_window(filepath):
    if system() == 'Darwin':  # macOS
        call(['open', filepath])
    elif system() == 'Windows':  # Windows
        os.startfile(filepath)
    elif system() == 'Linux':  # Linux
        call(['xdg-open', filepath])
    else:
        raise OSError('Unsupported operating system')
# to avoid breaking old versions. am repeating it here


def open_apsimx_file_in_window(filepath, bin_path):
    """
    Opens the given .apsimx file using the APSIM Next Generation executable located in `bin_path`.

    Parameters:
        filepath (str): Path to the .apsimx file.
        bin_path (str): Path to the APSIM binaries (folder containing ApsimNG or ApsimNG.exe).

    Raises:
        FileNotFoundError: If the APSIM executable does not exist.
    """
    if system() == 'Windows':
        executor = os.path.join(bin_path, 'ApsimNG.exe')
    else:
        executor = os.path.join(bin_path, 'ApsimNG')

    if not os.path.exists(executor):
        raise FileNotFoundError(f"APSIM executable not found at: {executor}")

    try:
        process = Popen([executor, filepath])
        return process  # Return the running process if needed
    except Exception as e:
        raise RuntimeError(f"Failed to open file: {e}")


class KeyValuePair:
    def __init__(self, Key, Value):
        self.key = Key
        self.value = Value


def generate_unique_name(base_name, length=6):
    random_suffix = ''.join(random.choices(string.ascii_lowercase, k=length))
    unique_name = base_name + '_' + random_suffix
    return unique_name


def delete_simulation_files(path, patterns=None):
    """

        :param path: path where the target files are located
        :param patterns: a list of different file patterns to delete
        :return: none
        """
    if not patterns:
        patterns = ['*lat*lon*.csv', 'clones*.apsimx', 'F*.apsimx', '*.db-shm', '*.db-wal', '*.db', 'fish*.cpg',
                    'edit*.bak', 'FB_*.db', 'coh*.bak', 'fish*.shp', 'Daymet*.met']
    localfiles = [glob.glob1(path, pat) for pat in patterns]
    files_to_delete = []
    for file_group in localfiles:
        absolute_paths = [opj(path, file) for file in file_group]
        files_to_delete.extend(absolute_paths)
    for i in files_to_delete:
        try:
            os.remove(i)
        except (FileNotFoundError, PermissionError) as e:
            # print('error occured while deleting file:', repr(e))
            pass


# _______________
def upload_weather(path, specific_number):
    pattern = f'daymet_wf_{specific_number}.met'
    localfiles = glob.glob1(path, pattern)
    if len(localfiles) == 0:
        logger.info(f"weather file for: {specific_number} not found")
    else:
        absolute_paths = [opj(path, file) for file in localfiles]
        return absolute_paths[0]


def upload_apsimx_file(path, specific_number):
    pattern = f'spatial_ap_{specific_number}.apsimx'
    localfiles = glob.glob1(path, pattern)
    if len(localfiles) == 0:
        logger.info(f"apsimx file for: {specific_number} not found")
    else:
        absolute_paths = [opj(path, file) for file in localfiles]
        return absolute_paths[0]


def upload_apsimx_file_by_pattern(path, specific_number, pattern):
    localfiles = glob.glob1(path, pattern)
    if len(localfiles) == 0:
        print(f"apsimx file for: {specific_number} not found")
    else:
        absolute_paths = [opj(path, file) for file in localfiles]
        return absolute_paths[0]


def make_apsimx_clones(base_file, number_of_clones):
    path = opj(os.getcwd(), 'apsimx_cloned')
    if os.path.exists(path):
        delete_simulation_files(path)
    if not os.path.exists(path):
        os.mkdir(path)
    files = []

    try:
        for i in range(number_of_clones):
            file_path = opj(path, generate_unique_name(f"clones_{i}")) + ".apsimx"
            files.append(file_path)
            shutil.copy(base_file, file_path)
        # check if there is any file that is not apsim
        for i in files:
            if not os.path.isfile(i) or not i.endswith('.apsimx'):
                files.remove(i)
                print(i, ": removed")
            yield i
    except Exception as e:
        print(repr(e))


def load_from_numpy(file):
    with open(file, 'rb') as f:
        ar = np.load(f, allow_pickle=True)
    return ar


def get_data_element(data, column_names, indexid):
    """_summary_

        Args:
            data (_array or dataframe_): _description_
            column_names (_type_): _columns name in n dimensional array
            
            indexid (_type_): index of the data to return

        Raises:
            ValueError: _erros if the column is not in the data

        Returns:
            data corresponding to the index id or position in the array
        """

    # Define a dictionary mapping statistic names to corresponding functions
    stat_functions = {
        'OBJECTID': lambda x: x[indexid],
        'FBndID': lambda x: x[indexid],
        'isAG': lambda x: x[indexid],
        'Shape': lambda x: x[indexid],
        'CropRotatn': lambda x: x[indexid]

    }

    # Get the desired statistic function from the dictionary
    func = stat_functions.get(column_names)

    if func is None:
        raise ValueError("Invalid column names. Supported are Shape, OBJECTID, FBndID, isAG and CropRotatn")

    # Calculate the desired statistic for the specified column
    result = func(data[column_names])

    return result


evaluate_crop = lambda x: not any(crop in x for crop in ['F', 'U', 'X', 'T', 'G', 'R', 'P', 'I'])
crop_mapping = {'C': 'Maize', 'B': 'Soybean', 'W': 'Wheat'}


def organize_crop_rotations(arrr):
    """_summary_

    Args:
        arrr (array_): _description_

    Returns:
        return concatenated crop rotation names
    """
    data = {}
    list_data = [list(y) for y in arrr]
    for i in range(len(list_data)):
        if all([evaluate_crop(crop) for crop in arrr[i]]):
            dt = ", ".join([crop_mapping[c] for c in arrr[i]])
            data[i] = dt
    return data


def split_and_replace(data):
    crop_mapping = {'C': 'Maize', 'B': 'Soybean', 'U': 'Unknown', 'P': 'Potato'}
    rows = data.strip().split('\n')
    result = []
    for row in rows:
        row_data = [crop_mapping[c] for c in row]
        result.append(", ".join(row_data))
        print(result)
    return "".join(result)


def convert_negative_to_positive(arr, column_index):
    # Check if all values in the specified column are negative
    for i in column_index:
        # Convert all negative values to positive
        arr[:, i] = np.abs(arr[:, i])
        print(f"changing column {i} to postive")
    return arr


def split_and_replace(data):
    """
    Replaces rows in the 'CropRotatn' column of the input structured array ('data')
    with calculated values based on a mapping ('crop_mapping').

    Parameters:
        data (numpy structured array): A structured array with a 'CropRotatn' column.

    Returns:
        numpy array: A new column containing calculated values for rows that do not meet the condition,
                     and 'none' for rows that meet the condition.
    """

    # Crop mapping dictionary to replace crop codes with crop names
    crop_mapping = {'C': 'Maize', 'B': 'Soybean', 'W': 'Wheat'}

    # Use np.vectorize() to apply np.char.find() on the specified columns
    find_func = np.vectorize(lambda x: not any(crop in x for crop in ['F', 'U', 'X', 'T', 'G', 'R', 'P', 'I']))
    valid_indices = np.where(find_func(data['CropRotatn']))[0]

    # Initialize a new column with 'none'
    new_column = np.array(['none'] * len(data), dtype=object)

    # Calculate and populate new_column for the rows that do not meet the condition
    new_column[valid_indices] = [", ".join([crop_mapping[c] for c in row]) for row in data['CropRotatn'][valid_indices]]

    return new_column


def convert_fc_to_numpy(fc):
    # Replace 'path_to_shapefile.shp' with the actual path to your feature class
    path_to_shapefile = fc

    # Read the shapefile into a GeoDataFrame
    gdf = gpd.read_file(path_to_shapefile)
    # Check the GeoDataFrame to ensure it was read correctly
    print(gdf.head())
    atn = ['FBndID', 'Acres', 'isAG', 'updateYr', 'Shape_Leng', 'Shape_Area']
    numpy_array = gdf[atn].to_numpy()

    # Now you have your attribute data as a NumPy array
    print(numpy_array)
    return numpy_array


def get_centroid(polygon):
    gdf = gpd.read_file(polygon)
    # Union all geometries in the GeoDataFrame
    unified_geometry = unary_union(gdf['geometry'])
    # Calculate centroid of the unified geometry
    centroid = unified_geometry.centroid
    centroid_gdf = gpd.GeoDataFrame(index=[0], crs=gdf.crs, geometry=[Point(centroid.x, centroid.y)])
    # Replace 'EPSG:4326' with the EPSG code of your target CRS
    reprojected_centroid_gdf = centroid_gdf.to_crs('EPSG:4326')
    lonlat = reprojected_centroid_gdf['geometry'].x[0], reprojected_centroid_gdf['geometry'].y[0]
    return lonlat


def create_polygon(lat, lon, lon_step, lat_step):
    return Polygon([(lon, lat), (lon + lon_step, lat), (lon + lon_step, lat + lat_step), (lon, lat + lat_step)])


def create_fishnet(min_lat, min_lon, max_lat, max_lon, lon_step, lat_step):
    lats = np.arange(min_lat, max_lat, lat_step)
    lons = np.arange(min_lon, max_lon, lon_step)

    polygons = []
    with ThreadPoolExecutor(max_workers=18) as executor:  # Adjust max_workers as needed
        x = 0
        for lon in lons:
            x += 1
            print(x)
            for lat in lats:
                print(x)
                future = executor.submit(create_polygon, lat, lon, lon_step, lat_step)
                polygons.append(future.result())

    gdf = gpd.GeoDataFrame({'geometry': polygons})


crs = "EPSG:4326"


def create_polygon1(args):
    lon, lat, lon_step, lat_step = args
    return Polygon([(lon, lat), (lon + lon_step, lat), (lon + lon_step, lat + lat_step), (lon, lat + lat_step)])


def create_fishnet1(pt, lon_step=20, lat_step=20, ncores=2, process=False):
    gdf_shape = gpd.read_file(pt)
    CRS = gdf_shape.crs
    print(crs)
    min_lon, min_lat, max_lon, max_lat = gdf_shape.total_bounds
    lats = np.arange(min_lat, max_lat, lat_step)
    lons = np.arange(min_lon, max_lon, lon_step)
    polygons = []

    with select_process(process, ncores) as executor:
        args = [(lon, lat, lon_step, lat_step) for lon in lons for lat in lats]
        polygons = list(executor.map(create_polygon1, args))
    gdf = gpd.GeoDataFrame({'geometry': polygons}, crs=CRS)
    gdf_clip = gpd.clip(gdf, gdf_shape)

    return gdf_clip


def convert_geopoint_to_array(geoseries):
    # Assuming 'geoseries' is your GeoSeries of Point geometries

    # Extract the x and y coordinates
    x_coords = geoseries.geometry.x
    y_coords = geoseries.geometry.y

    # Stack the coordinates into a 2D array
    coords_array = np.column_stack((x_coords, y_coords))

    return coords_array


import numpy as np
import math


def split_arr(ar, chunk_size):
    num_chunks = math.ceil(len(ar) / chunk_size)
    indices = range(chunk_size, len(ar), num_chunks)
    return np.vsplit(ar, indices)


def moving_average(iterable, n=3):
    import itertools
    # moving_average([40, 30, 50, 46, 39, 44]) --> 40.0 42.0 45.0 43.0
    # https://en.wikipedia.org/wiki/Moving_average
    it = iter(iterable)
    d = deque(itertools.islice(it, n - 1))
    d.appendleft(0)
    s = sum(d)
    for elem in it:
        s += elem - d.popleft()
        d.append(elem)
        yield s / n


def Cache(func):
    """

    This is a function decorator for class attributes. It just remembers the result of the FIRST function call
    and returns this from there on.
    """

    func_name = func.__name__

    def wrapper(self, *args, use_cache=True, set_cache=True, **kwargs):

        if "cache" not in self.__dict__:
            self.__dict__["cache"] = {}

        cache = self.__dict__["cache"]

        if use_cache and func_name in cache:
            return cache[func_name]
        else:

            obj = func(self, *args, **kwargs)

            if set_cache:
                cache[func_name] = obj

            return obj

    return wrapper


def add_wheat(string_object, word_to_insert="Wheat"):
    result = []
    words = string_object.split(', ')
    last_target_index = -1
    target_words = [words[0]] + ["Soybean"]
    ["Maize", "Soybean"]
    # Iterate through the words in the list
    for i, word in enumerate(words):
        result.append(word)  # Add the current word to the result list
        # Check if the current word is one of the target words
        if word in target_words and len(
                words) - 1 > i > last_target_index and word_to_insert not in string_object:
            result.append(word_to_insert)  # Add the word to insert after the target word
            last_target_index = i  # Update the last target word index
    return result


def assign_management_practices(locations, object_ids, num_practices):
    if len(locations) != len(object_ids):
        raise ValueError("The 'locations' and 'object_ids' arrays must have the same length.")
    watershed_data = []
    for location, obj_id in zip(locations, object_ids):
        # Generate a random management practice ID between 1 and num_practices
        management_practice_id = random.randint(1, num_practices)
        # Append the location, objectID, and management practice ID to the watershed_data
        watershed_data.append({"location": location, "objectID": obj_id, "practiceID": management_practice_id})
    return watershed_data


def idex_excutor(self, x):  # We supply x from the rotations idex which inherits objectid
    try:
        a = time.perf_counter()
        print(f"downloading for: {x}")
        data_dic = {}
        fn = "daymet_wf_" + str(x) + '.met'
        cod = pol_object.record_array["Shape"][x]

        filex = weather.daymet_bylocation_nocsv(cod, start=1998, end=2020, cleanup=False, filename=fn)
        return filex
    except Exception as e:
        # raise
        print(e, "has occured")
        try:
            print("trying again")
            filex = weather.daymet_bylocation_nocsv(cod, start=1998, end=2000, cleanup=True, filename=fn)
            return filex
        except:
            print("unresolved errors at the momentt try again")


def threaded_weather_download(self, iter_arable):
    threads = []
    for idices in iter_arable:
        thread = Thread(target=self.idex_excutor, args=(idices,))
        threads.append(thread)
        thread.start()
    # Wait for all threads to finish
    for thread in threads:
        thread.join()


# # Example usage:
# locations_array = [(1, 2), (3, 4), (5, 6), (7, 8), (1, 2), (3, 4), (5, 6), (7, 8)]  # Example array of locations (x, y coordinates)
# object_ids_array = [101, 102, 103, 104, 105, 106, 107, 109]  # Example array of object IDs
# num_management_practices = 5  # Total number of different management practices

# result_data = assign_management_practices(locations_array, object_ids_array, num_management_practices)
#
# print(pd.DataFrame(result_data))
# some ideas
# __________________________________________________________________________
# we can constrain the sample size of each management practice every time we sample
def collect_runfiles(path2files, pattern="*.apsimx"):
    """_summary_

    Args:
        path2files (_type_): path to the apsimx or database file_
        pattern (str, optional) or lists of strings: file pattern. Defaults to "*.apsimx".

    Returns:
        _type_: _description_
    """
    lp = []
    if not isinstance(pattern, list):
        pattern = [pattern]
    for i in pattern:
        list_files = glob.glob1(path2files, i)
        lp += list_files
    files = [os.path.join(path2files, i) for i in lp]
    assert len(lp) != [], "No files found please try another path or file pattern"
    for i in files:
        yield i


def decreasing_exponential_function(x, a, b):  # has potential to cythonize
    """
    Compute the decreasing exponential function y = a * e^(-b * x).

    Parameters:
        x (array-like): Input values.
        a (float): Amplitude or scaling factor.
        b (float): Exponential rate.

    Returns:
        numpy.ndarray: The computed decreasing exponential values.
    """
    func = a * np.exp(-b * x)
    return func


def optimize_exponetial_data(x_data, y_data, initial_guess=[0.5, 0.5],
                             bounds=([0.1, 0.01], [np.inf, np.inf])):  # defaults for carbon

    best_fit_params, _ = curve_fit(decreasing_exponential_function, x_data, y_data, p0=initial_guess,
                                   bounds=bounds)
    a_fit, b_fit = best_fit_params
    predicted = decreasing_exponential_function(x_data, a_fit, b_fit)
    return predicted


def area_of_circle(r):
    pi = 3.141592653589793
    return pi * r * r


def number_of_cells(r, cell_size):
    circle_area = area_of_circle(r)
    cell_area = cell_size * cell_size
    return int(circle_area / cell_area)


def timer(func):
    def wrapper(*args, **kwargs):
        start_time = perf_counter()
        result = func(*args, **kwargs)
        end_time = perf_counter()
        elapsed_time = end_time - start_time
        logger.info(f"{func.__name__} took {elapsed_time:.4f} seconds to execute.")
        return result

    return wrapper


def filter_df(df, **kwargs):
    """
    Filter a DataFrame based on values in specified columns.

    Args:
        df (pd.DataFrame): The DataFrame to be filtered.
        **kwargs: Keyword arguments where the key is the column name and the value is the value to filter on.

    Returns:
        pd.DataFrame: The filtered DataFrame.

    Example:
        filtered_df = filter_dataframe(df, Age=30, City='Los Angeles')
    """
    # Initialize a boolean mask with True values
    mask = pd.Series(True, index=df.index)

    # Iterate through keyword arguments and update the mask for each column and value pair
    for column, value in kwargs.items():
        mask &= (df[column] == value)

    # Apply the mask to the DataFrame to filter the rows
    filtered_df = df[mask]
    return filtered_df


@cache
def find_models(path, filename):
    mod = list(path.rglob(filename))
    if mod != []:
        return os.path.dirname(mod[0])
    else:
        return None


def convert_df_to_gdf(df, CRS):
    """
    Converts a pandas DataFrame to a GeoPandas GeoDataFrame with specified CRS.

    This function assumes that the input DataFrame contains a column named 'geometry'
    with WKT (Well-Known Text) representations of geometric objects. It converts these
    textual geometry representations into GeoPandas geometry objects and sets the
    coordinate reference system (CRS) for the resulting GeoDataFrame.

    Parameters:
    - df (pandas.DataFrame): The input DataFrame to convert. Must include a 'geometry'
      column with WKT representations of the geometries.
    - CRS (str or dict): The coordinate reference system to set for the GeoDataFrame.
      Can be specified as a Proj4 string, a dictionary of Proj parameters, an EPSG
      code, or a WKT string.

    Returns:
    - gpd.GeoDataFrame: A GeoDataFrame created from the input DataFrame with geometries
      converted from WKT format and the specified CRS.

    Example:
    ```
    import pandas as pd
    import geopandas as gpd
    from shapely import wkt

    # Sample DataFrame with WKT geometries
    data = {'geometry': ['POINT (12 34)', 'LINESTRING (0 0, 1 1, 2 1, 2 2)'],
            'value': [1, 2]}
    df = pd.DataFrame(data)

    # Convert DataFrame to GeoDataFrame with an EPSG code for CRS
    gdf = convert_df_to_gdf(df, CRS='EPSG:4326')
    ```

    Note:
    The 'wkt' module from 'shapely' needs to be imported to convert geometries from WKT.
    """
    df['geometry'] = df['geometry'].apply(wkt.loads)
    gdf = gpd.GeoDataFrame(df, crs=CRS)
    return gdf


def bounding_box_corners(center_point, radius):
    """
        Generates a bounding box around a center point within a specified radius and returns all four corners.

        Parameters:
        - center_point: tuple of (latitude, longitude)
        - radius: radius in meters

        Returns:
        - A dictionary with the coordinates of the 'north_east', 'south_east', 'south_west', and 'north_west' corners of the bounding box.
        """

    # Calculate the north, south, east, and west points by applying the radius distance to the center point
    north = geodesic(meters=radius).destination(center_point, bearing=0)
    south = geodesic(meters=radius).destination(center_point, bearing=180)
    east = geodesic(meters=radius).destination(center_point, bearing=90)
    west = geodesic(meters=radius).destination(center_point, bearing=270)

    # Calculate corners by combining the north/south with east/west
    # we wanted to be intuitive in the direction calculation but let' reverse it'
    north_east = (east.longitude, north.latitude)
    south_east = (east.longitude, south.latitude)
    south_west = (west.longitude, south.latitude)
    north_west = (west.longitude, north.latitude)

    ne = north_east
    sw = south_east
    se = south_west
    nw = north_west
    return ne, se, sw, nw


class WDir:
    def __init__(self, path_dir=None):
        assert path_dir, "path dir_path is required"
        self.initial_path = Path(path_dir)

    def path(self, name=None):
        """

        :param name: name of the new file
        :return: realpath for the new file name
        """
        assert name, "name is required"
        return os.path.realpath(self.initial_path.joinpath(name))

    def mkdir(self, name):
        new_dir = self.initial_path.joinpath(name)
        new_dir.mkdir(exist_ok=True)
        return new_dir

    def make_this_cwd(self):
        wd = os.path.realpath(self.initial_path)
        if os.getcwd() != wd:
            os.chdir(os.path.realpath(self.initial_path))
        else:
            logger.warning(f"this path: {wd} is already the current working dir_path")


def exception_handler(re_raise=False):
    """A decorator to handle exceptions with an option to re-raise them.

    Args:
    re_raise (bool): If True, re-raises the caught exception after logging.
                     If False, logs the exception and suppresses it.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                # Extract unformatted stack traces as a tuple
                trace_back = traceback.extract_tb(exc_traceback)
                # Format stacktrace
                stack_trace = list()

                for trace in trace_back:
                    stack_trace.append("File : %s , Line : %d, Func.Name : %s, Message : %s" % (
                        trace[0], trace[1], trace[2], trace[3]))

                logger.error(f"Exception occurred in {func.__name__}: {e}")
                print("Error details : ")
                logger.info("\n".join(stack_trace))

                if re_raise:
                    raise

                return None

        return wrapper

    return decorator


def flatten_nested_list(nested_list, deep=True):
    """
    this will recursively flatten a nested list
    :param nested_list:  to flatten
    :return: a list with nested flattened values
    """
    flattened = []
    for item in nested_list:

        if isinstance(item, (tuple, list, set)):
            # Recursively flatten the sublist
            flattened.extend(flatten_nested_list(item))
        else:
            # If the item is not a list, append it to the flattened list
            flattened.append(item)
    return flattened


def flatten_dict(nested_dict, parent_key='', separator='.'):
    flattened = {}
    for key, value in nested_dict.items():
        new_key = f"{parent_key}{separator}{key}" if parent_key else key
        if isinstance(value, dict):
            # Recursively flatten the sub-dictionary
            flattened.update(flatten_dict(value, new_key, separator=separator))
        else:
            # Add the current key-value pair to the flattened dictionary
            flattened[new_key] = value
    return flattened


if __name__ == '__main__':
    ...



from concurrent.futures import as_completed

import geopandas as gpd
import pandas as pd
import numpy as np
from apsimNGpy.core import ApsimModel
from apsimNGpy.core_utils.utils import select_process
from shapely.geometry import Polygon
import random
from shapely.geometry import Point
import os
from shutil import copy
from shapely.ops import unary_union
from apsimNGpy.core.base_data import LoadExampleFiles
from apsimNGpy.parallel.process import download_soil_tables
from tqdm import tqdm
import random
from apsimNGpy.manager.soilmanager import OrganiseSoilProfile, DownloadsurgoSoiltables
from apsimNGpy.core.weather import daymet_bylocation_nocsv
from apsimNGpy.parallel.process import custom_parallel
from apsimNGpy.parallel.safe import initialise

maize = LoadExampleFiles().get_maize

WGS84 = 'epsg:4326'

shp = r'D:\ACPd\Bear creek simulations\bearcreek_shape\bearcreek.shp'


def create_polygon1(args):
    lon, lat, lon_step, lat_step = args
    return Polygon([(lon, lat), (lon + lon_step, lat), (lon + lon_step, lat + lat_step), (lon, lat + lat_step)])


def create_fishnet1(pt, lon_step=200, lat_step=200, ncores=3, use_thread=True, **kwargs):
    """

    Args: pt: shape or point feature class layer lon_step: height of the polygon lat_step: width of the polygon
    ncores: number of cores to use use_thread: if True, threads will be used if false processes will be used
    **kwargs: use key word Return = gdf to return GeoPandas data frame: this is show polygon coordinates otherwise if
    not supplied to will returun an array

    Returns: an array or geopandas data frame
    """
    gdf_shape = gpd.read_file(pt)
    CRS = gdf_shape.crs
    print(CRS)
    min_lon, min_lat, max_lon, max_lat = gdf_shape.total_bounds
    lats = np.arange(min_lat, max_lat, lat_step)
    lons = np.arange(min_lon, max_lon, lon_step)
    polygons = []

    with select_process(use_thread, ncores) as executor:
        args = [(lon, lat, lon_step, lat_step) for lon in lons for lat in lats]
        polygons = list(executor.map(create_polygon1, args))
    gdf = gpd.GeoDataFrame({'geometry': polygons}, crs=CRS)
    gdf_clip = gpd.clip(gdf, gdf_shape)
    gdf = gdf_clip.centroid
    gdf_transformed = gdf.to_crs(WGS84)
    if kwargs.get("Return", 'array') == 'array':
        return np.array([(point.x, point.y) for point in gdf_transformed])
    else:
        return gdf_clip


# Function to generate random points within a polygon
def generate_random_points(pt, resolution, ncores, num_points):
    """

    Args:
        pt: shape file
        resolution: resolution in meters
        ncores: number of cores to use
        num_points: number of points to sample in each grid

    Returns:

    """
    all_points = []

    def generate():
        min_x, min_y, max_x, max_y = poly.bounds
        points = []

        while len(points) < num_points:
            random_point = Point([random.uniform(min_x, max_x), random.uniform(min_y, max_y)])
            if random_point.within(poly):
                points.append(random_point)
        return points

    gdf = create_fishnet1(pt, lon_step=resolution, lat_step=resolution, ncores=ncores, Return='gdf')
    df = gdf.to_crs(WGS84)
    for poly in df['geometry']:
        random_points = generate()  # Generate 3 random points
        all_points.extend(random_points)
        points_gdf = gpd.GeoDataFrame(geometry=all_points)  # Add the points to the list
        GDF = points_gdf
    return np.array([(point.x, point.y) for point in GDF.geometry])


def random_points_in_polygon(number, polygon, seed=1):
    random.seed(seed)
    points = []
    min_x, min_y, max_x, max_y = polygon.bounds
    i = 0
    while i < number:
        point = Point(random.uniform(min_x, max_x), random.uniform(min_y, max_y))
        if polygon.contains(point):
            points.append(point)
            i += 1
    return points


def samply_by_polygons(shp, k=1, filter_by=None, filter_value=None):
    points = []

    gd = gpd.read_file(shp)
    total_area = gd.area.sum() / 10 ** 6
    print(f"The area is {total_area}")
    if filter_by:
        gd = gd[gd[f"{filter_by}"] == filter_value]
    CRS = gd.crs
    import math
    poi = []
    for poly in gd['geometry']:
        area = poly.area / 10 ** 6
        weight = area / total_area
        new_k = math.ceil(area * k)
        poi.append(new_k)
        print(new_k)
        pts = random_points_in_polygon(new_k, poly)
        points.extend(pts)
        points_gdf = gpd.GeoDataFrame(geometry=points, crs=CRS)
        gdf = points_gdf.to_crs(WGS84)
        # gdf['lonlats'] =np.array([(point.x, point.y) for point in gdf.geometry])
    print(np.sum(poi))
    return np.array([(point.x, point.y) for point in gdf.geometry])


def create_apsimx_sim_files(wd, model, iterable):
    """
    Creates copies of a specified APSIM model file for each element in the provided iterable,
    renaming the files to have unique identifiers based on their index in the iterable.
    The new files are saved in the specified working dir_path.

    Args:
    wd (str): The working dir_path where the new .apsimx files will be stored.
    model (str): The path to the .apsimx model file that will be copied.
    iterable (iterable): An iterable (e.g., list or range) whose length determines the number of copies made.

    Returns:
    dict: A dictionary where keys are indices from 0 to len(iterable)-1 and values are paths to the newly created .apsimx files.

    The function performs the following steps:
    1. Extracts the basename of the model file, removing the '.apsimx' extension to create a model suffix.
    2. Iterates over the `iterable`, creating a unique file name for each element by appending an index and '.apsimx' to the model suffix.
    3. Copies the original model file to the new file name in the specified working dir_path.
    4. Returns a dictionary mapping each index to the file path of the created .apsimx file.

    Example:
    >>wd = '/path/to/working/dir_path'
    >> model = '/path/to/original/model.apsimx'
    >> file_paths = create_apsimx_files(wd, model, range(5))
    >> print(file_paths)
    {0: '/path/to/working/dir_path/model_0.apsimx', 1: '/path/to/working/dir_path/model_1.apsimx', ...}
    """

    mod = model.strip('.apsimx')
    model_suffix = os.path.basename(mod)

    ids = range(len(iterable))
    files = [{'ID': i, 'location': iterable[i], 'file_name': os.path.join(wd, model_suffix) + f"_{i}.apsimx"} for i in
             ids]
    df = pd.DataFrame(files)
    [copy(model, df['file_name'][i]) for i in ids]
    return df


def download_weather(df, start, end, use_thread=True, ncores=10, replace_soils=True, **kwargs):
    """
       downloads and replace soil or weather files or both in parallel or threads
    Args:
        replace_soils: Set this to true to simulataneoursly downloand and replace soils
        df: data frame generated by 'create_apsimx_sim_files'
        start: start year of the simulation
        end:  end year of the simulation
        use_thread: if true threading will take place otherwise multiprocessing
        ncores: number of cores to use
    kwargs:
      verbose: bool, Set to True print current step
      thickness_values: list defining the soil layer thickness
      report : set to true to return results
      report_names; provide the required table names from apsimx model report
    Returns:

    """

    def worker(row):
        model = str(row['file_name']).strip('.apsimx')
        ID = row['ID']
        location = row['location']
        wname = f"{model}_{ID}.met"
        out_path_name = f"{model}_{ID}.apsimx"
        model = row['file_name']
        thi = [150, 150, 200, 200, 200, 250, 300, 300, 400, 500]
        th = kwargs.get("thickness_values", thi)
        mod = ApsimModel(model, thickness_values=th, out_path=out_path_name)
        if kwargs.get('replace_weather', True):
            wf = daymet_bylocation_nocsv(location, start=start, end=end, filename=wname)
            sim_name = mod.extract_simulation_name
            mod.replace_met_file(wf, sim_name)
        if kwargs.get("verbose"):
            print("downloading and replacing soils now")
        if replace_soils:
            table = DownloadsurgoSoiltables(location)
            if isinstance(table, pd.DataFrame):
                sp = OrganiseSoilProfile(table, thickness=20, thickness_values=th)
                sp = sp.cal_missingFromSurgo()
                if sp[0].isna().any().any() or sp[1].isna().any().any() or sp[2].isna().any().any():
                    print(f"soils not replaced at {location}")
                    return None
                else:
                    mod.replace_downloaded_soils(sp, sim_name)
        if kwargs.get("report"):
            mod.run(report_name=kwargs.get('report_names'))
            return mod.results
        else:
            mod.save_edited_file(out_path_name)
            mod.clear()
            del mod
            return out_path_name

    with select_process(use_thread, ncores) as tpool:
        futures = {tpool.submit(worker, df.loc[df['ID'] == i].squeeze()): i for i in df['ID']}
        progress = tqdm(total=len(futures), position=0, leave=True,
                        bar_format=f'Running:' '{percentage:3.0f}% completed')
        # Iterate over the futures as they complete
        for future in as_completed(futures):
            yield future.result()
            progress.update(1)
        progress.close()
        # for future in as_completed(futures):


def create_and_run_sim_objects(wd, shp_file, resolution, num_points, model_file, reports_names, cores=10, **kwargs):
    """

    Args:
        wd: working dir_path
        shp_file: shape file of the target area
        model_file: APSIM model string path
        reports_names:str or list names of the data in the simulation model
        **kwargs:
           Test: bool. set to true to try out 10 sample before simulation
           run_process: set too false to run in parallel
           select_process; set too False to use multiple process
           'replace_weather'

    :param shp_file:
    :param resolution:int. square qrid resolution
    :param num_points:int for random sampling

    """
    print(f"Here is your extra arguments: {kwargs}")
    ap = generate_random_points(shp_file, resolution, 10, num_points)
    if kwargs.get('test'):
        k = 10
        random_indices = np.random.choice(ap.shape[0], size=k, replace=False)

        # Use the random indices to select rows from ap
        arr = ap[random_indices]
    else:
        arr = ap
    print("Downloading weather files")
    ap = create_apsimx_sim_files(wd, model_file, arr)
    # first we replace soils, then we
    objs = download_weather(ap, 1990, 2021, report_names='Carbon', ncores=cores, verbose=False, report=False,
                            use_thread=kwargs.get('select_process', True),
                            replace_soils=False, replace_weather=True)

    weathers = list(objs)

    print("\nDownloading soils now please wait")
    objs = download_weather(ap, 1990, 2021, report_names='Carbon', verbose=False, ncores=cores, report=False,
                            use_thread=kwargs.get('select_process', True),
                            replace_soils=True)
    soils = list(objs)
    data = [s for s in soils if s in weathers and s is not None]
    print(f"\nrunning the {len(data)} simulations now.....")
    sim = custom_parallel(initialise, data, reports_names, ncores=cores, use_thread=kwargs.get('run_process', True))
    return list(sim)


if __name__ == '__main__':
    wd = r'C:\Users\rmagala'
    fb401 = r'D:\ACPd\Long deek401 simulations\FB401.shp'
    df = create_fishnet1(shp, ncores=10, use_thread=True)
    gd = df
    bc_model = r'D:\ACPd\Bear creek simulations\ML_bear_creek 20240206.apsimx'
    fb = gpd.read_file(fb401)
    lp = samply_by_polygons(fb401, k=4, filter_by='isAG', filter_value=1)
    # data = create_and_run_sim_objects(wd, shp, 500, 2, maize, 'Carbon', test=False, run_process=False,
    #                                   select_process=True, cores=13)
    # # sims = run_created_files(data, "Carbon", cores = 15, use_threads = False)
    # # dat = custom_parallel(run_simPle, data, "Carbon", ncores=14, use_thread=True)
    # # dd = list(dat)
    # from joblib import dump, load
    #
    # # dat = load('sims')
    # # ap = ApsimModel(data[8])
    # # dump(data, 'sims')
    # # mod = [model.run(report_name='Carbon') for model in mop]



from dataclasses import dataclass, field
from apsimNGpy.core.apsim import ApsimModel
from scipy.optimize import minimize
import numpy as np
from functools import reduce, lru_cache
from collections import OrderedDict
from functools import wraps
from collections import OrderedDict
from abc import ABC, abstractmethod
from wrapdisc.var import ChoiceVar, GridVar, QrandintVar, QuniformVar, RandintVar, UniformVar
import inspect
from typing import Union
import wrapdisc



@lru_cache(maxsize=None)
def get_function_param_names(func):
    """
    Return only the parameter names of a function as a tuple.
    """
    sig = inspect.signature(func)
    return tuple(sig.parameters.keys())


class AbstractProblem(ApsimModel):
    def __init__(self, model, max_cache_size=None):
        super().__init__(model)
        self._cache = OrderedDict()
        self.max_cache_size = max_cache_size
        self._outcomes = None
        self.labels = []

    def _insert_cache_result(self, *args, result):

        args  =tuple(args)
        self._cache[args] = result

        if self.max_cache_size is not None and len(self._cache) > self.max_cache_size: # saturation reached according to specified maxsize free space
            self._cache.popitem(last=False)

    def get_cached(self, *args):
        key = tuple(args)
        return self._cache.get(key, None)

    def clear_cache(self):
        self._cache.clear()
    @abstractmethod
    def add_control(self):
        pass
    @abstractmethod
    def evaluate(self, x):
        pass

    @abstractmethod
    def minimize_with_local_solver(self, **kwargs):
        pass

    def __str__(self):
        summary = f"Optimization Problem for APSIM Model: {self.model}\n"
        summary += f"===================================================\n"
        summary += f"Simulation: {self.simulation}\n"
        summary += f"Number of Control Variables: {len(self.control_vars)}\n\n"
        summary += "Control Variables:\n"
        for var in self.control_vars:
            summary += (
                f"  - {var.label} ({type(var.vtype).__name__}): "
                f"model_type={var.model_type}, model_name='{var.model_name}', "
                f"param='{var.parameter_name}'\n"
                f"Results summary:: {self.outcomes}\n"
            )
        return summary

    def __repr__(self):
        return self.__str__()

    @property
    def outcomes(self):
        return self._outcomes

    @outcomes.setter
    def outcomes(self, value):
        self._outcomes = value

    def _evaluate_args(self, *args, **kwargs):
        if not all(isinstance(arg, str) for arg in args):
            raise ValueError("all arguments must be strings")
    def _auto_guess(self, data):
        if isinstance(data, ChoiceVar):
            sample_set = np.random.choice(data.categories, size=1)[0]
        elif isinstance(data, GridVar):
            sample_set = np.random.choice(data.values, size=1)[0]
        elif isinstance(data, (QrandintVar, RandintVar, QuniformVar)):
            sample_set = data.lower
        elif isinstance(data, UniformVar):
            if len(data.bounds) == 1:
                bounds = data.bounds[0]
            else:
                bounds = data.bounds
            sample_set = np.random.uniform(bounds[0], bounds[1], size=1)[0]
        else:
            raise ValueError(f'data: {type(data)} not supported')
        return sample_set

@dataclass(slots=True, frozen=True)
class VarDesc:
    model_type: str
    model_name: str
    parameter_name: str
    vtype: Union[str, object, wrapdisc.var]
    label: str
    start_value: Union[float, int]
    bounds: tuple[float, int]






import unittest
from apsimNGpy.tests.tester_main import suite, loader, run_suite
from apsimNGpy.core.base_data import load_default_simulations


class TestCaseAddModule(unittest.TestCase):

    def setUp(self):
        self.model = load_default_simulations('Maize')
        self.out = 'test_edit_model.apsimx'

    def test_add_crop_replacement(self):
        "+++test adding crop replacement++"
        self.model.add_crop_replacements(_crop='Maize')
        self.model.create_experiment(permutation=True)


if __name__ == '__main__':
    suite.addTests(loader.loadTestsFromTestCase(TestCaseAddModule))
    run_suite(2)



import requests
import pandas as pd
import xmltodict
def get_gssurgo_soil_soil_table_at_lonlat(lonlat, select_componentname=None, summarytable=False):
    '''
    TODO this is a duplicate File. Duplicate of soils/soilmanager
    Downloads SSURGO soil tables

    parameters
    ------------------
    lon: longitude 
    lat: latitude
    select_componentname: any componet name within the map unit e.g 'Clarion'. the default is None that mean sa ll the soil componets intersecting a given locationw il be returned
      if specified only that soil component table will be returned. in case it is not found the dominant componet will be returned with a caveat meassage.
        use select_componentname = 'domtcp' to return the dorminant component
    summarytable: prints the component names, their percentages
    '''

    total_steps = 3
    # lat = "37.54189"
    # lon = "-120.96683"
    lonLat = "{0} {1}".format(lonlat[0], lonlat[1])
    url = "https://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx"
    # headers = {'content-child': 'application/soap+xml'}
    headers = {'content-child': 'text/xml'}
    body = """<?xml version="1.0" encoding="utf-8"?>
              <soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:sdm="http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx">
       <soap:Header/>
       <soap:Body>
          <sdm:RunQuery>
             <sdm:Query>SELECT co.cokey as cokey, ch.chkey as chkey, comppct_r as prcent, compkind as compkind_series, wsatiated_r as wat_r,partdensity as pd, dbthirdbar_h as bb, musym as musymbol, compname as componentname, muname as muname, slope_r, slope_h as slope, hzname, hzdept_r as topdepth, hzdepb_r as bottomdepth, awc_r as PAW, ksat_l as KSAT,
                        claytotal_r as clay, silttotal_r as silt, sandtotal_r as sand, texcl, drainagecl, om_r as OM, iacornsr as CSR, dbthirdbar_r as BD, wfifteenbar_r as L15, wthirdbar_h as DUL, ph1to1h2o_r as pH, ksat_r as sat_hidric_cond,
                        (dbthirdbar_r-wthirdbar_r)/100 as bd FROM sacatalog sc
                        FULL OUTER JOIN legend lg  ON sc.areasymbol=lg.areasymbol
                        FULL OUTER JOIN mapunit mu ON lg.lkey=mu.lkey
                        FULL OUTER JOIN component co ON mu.mukey=co.mukey
                        FULL OUTER JOIN chorizon ch ON co.cokey=ch.cokey
                        FULL OUTER JOIN chtexturegrp ctg ON ch.chkey=ctg.chkey
                        FULL OUTER JOIN chtexture ct ON ctg.chtgkey=ct.chtgkey
                        FULL OUTER JOIN copmgrp pmg ON co.cokey=pmg.cokey
                        FULL OUTER JOIN corestrictions rt ON co.cokey=rt.cokey
                        WHERE mu.mukey IN (SELECT * from SDA_Get_Mukey_from_intersection_with_WktWgs84('point(""" + lonLat + """)')) 

                        AND sc.areasymbol != 'US' 
                        order by co.cokey, ch.chkey, prcent, topdepth, bottomdepth, muname
            </sdm:Query>
          </sdm:RunQuery>
       </soap:Body>
    </soap:Envelope>"""

    response = requests.post(url, data=body, headers=headers, timeout=140)
    # Put query results in dictionary format
    my_dict = xmltodict.parse(response.content)

    # Convert from dictionary to dataframe format
    soil_df = None
    try:
        soil_df = pd.DataFrame.from_dict(
            my_dict['soap:Envelope']['soap:Body']['RunQueryResponse']['RunQueryResult']['diffgr:diffgram'][
                'NewDataSet'][
                'Table'])
    except ValueError as e:
        pass
    if isinstance(soil_df, pd.DataFrame):
        if summarytable:
            df = soil_df.drop_duplicates(subset=['componentname'])
            summarytable = df[["componentname", 'prcent', 'chkey']]
            print("summary of the returned soil tables \n")
            print(summarytable)
        # select the dominat componet
        dom_component = soil_df[soil_df.prcent == soil_df.prcent.max()]
        # select by component name
        if select_componentname in soil_df.componentname.unique():
            componentdf = soil_df[soil_df.componentname == select_componentname]
            return componentdf
        elif select_componentname == 'domtcp':
            return dom_component
            # print("the following{0} soil components were found". format(list(soil_df.componentname.unique())))
        elif select_componentname is None:
            # print("the following{0} soil components were found". format(list(soil_df.componentname.unique())))
            return soil_df
        elif select_componentname != 'domtcp' and select_componentname not in soil_df.componentname.unique() or select_componentname != None:
            print(
                f'Ooops! we realised that your component request: {select_componentname} does not exists at the '
                f'specified location. We have returned the dorminant component name')
            return dom_component


def get_gssurgo_soil_soil_table_at_polygon(_coordinates):
    """
        Function to get al soil mapunit keys or components intersection a polygon
    :param _coordinates: a list returned by get_polygon_bounds function
    :return: data frame
    """
    formatted_coordinates = ",".join([f"{lon} {lat}" for lon, lat in _coordinates])
    url = "https://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx"
    headers = {'content-child': 'application/soap+xml'}
    body = f"""<?xml version="1.0" encoding="utf-8"?>
                    <soap:Envelope xmlns:soap="http://www.w3.org/2003/05/soap-envelope" xmlns:sdm="http://SDMDataAccess.nrcs.usda.gov/Tabular/SDMTabularService.asmx">
             <soap:Header/>
             <soap:Body>
                <sdm:RunQuery>
                   <sdm:Query>SELECT co.cokey as cokey, ch.chkey as chkey, comppct_r as prcent, compkind as compkind_series, lg.areasymbol, wsatiated_r as wat_r,partdensity as pd, dbthirdbar_h as bb, musym as musymbol, compname as componentname, muname as muname, slope_r, slope_h as slope, hzname, hzdept_r as topdepth, hzdepb_r as bottomdepth, awc_r as PAW, ksat_l as KSAT,
                              claytotal_r as clay, silttotal_r as silt, sandtotal_r as sand, texcl, drainagecl, om_r as OM, iacornsr as CSR, dbthirdbar_r as BD, wfifteenbar_r as L15, wthirdbar_h as DUL, ph1to1h2o_r as pH, ksat_r as sat_hidric_cond,
                              (dbthirdbar_r-wthirdbar_r)/100 as bd FROM sacatalog sc
                              FULL OUTER JOIN legend lg  ON sc.areasymbol=lg.areasymbol
                              FULL OUTER JOIN mapunit mu ON lg.lkey=mu.lkey
                              FULL OUTER JOIN component co ON mu.mukey=co.mukey
                              FULL OUTER JOIN chorizon ch ON co.cokey=ch.cokey
                              FULL OUTER JOIN chtexturegrp ctg ON ch.chkey=ctg.chkey
                              FULL OUTER JOIN chtexture ct ON ctg.chtgkey=ct.chtgkey
                              FULL OUTER JOIN copmgrp pmg ON co.cokey=pmg.cokey
                              FULL OUTER JOIN corestrictions rt ON co.cokey=rt.cokey
                              WHERE mu.mukey IN (
                            SELECT * FROM SDA_Get_Mukey_from_intersection_with_WktWgs84('polygon((
                           {formatted_coordinates}))')
                   ) 
                              AND sc.areasymbol != 'US' 
                              order by co.cokey, ch.chkey, prcent, topdepth, bottomdepth, muname
                  </sdm:Query>
                </sdm:RunQuery>
             </soap:Body>
          </soap:Envelope>"""

    response = requests.post(url, data=body, headers=headers, timeout=140)
    # Put query results in dictionary format
    my_dict = xmltodict.parse(response.content)

    # Convert from dictionary to dataframe format
    soil_df = None
    try:
        soil_df = pd.DataFrame.from_dict(
            my_dict['soap:Envelope']['soap:Body']['RunQueryResponse']['RunQueryResult']['diffgr:diffgram'][
                'NewDataSet'][
                'Table'])
    except ValueError as e:
        pass
    return soil_df

def get_polygon_points(row):
    """Extract points from a polygon geometry."""
    if row.geometry.geom_type == 'Polygon':
        return list(row.geometry.exterior.coords)
    elif row.geometry.geom_type == 'MultiPolygon':
        points = []
        for polygon in row.geometry.geoms:
            points.extend(list(polygon.exterior.coords))
        return points
    else:
        return None  # not sure of other polygon types

# Apply the function to each row in the GeoDataFrame
def get_polygon_bounds(gdf_):
    gdf = gdf_.copy()
    crs = gdf.crs
    gdf.to_crs(4326, inplace=True)
    #print(f"coordinate system changed from: {crs} to: {gdf.crs}")
    return gdf.apply(get_polygon_points, axis=1)


# finally solved this problem
"""
This module will replace all other modules they are too complicated for nothing
"""
from apsimNGpy.core.core import CoreModel, Models, RENAME


class Experiment(CoreModel):
    def __init__(self, model, out_path=None, permutation: bool = True, base_model_simulation: str = None, **kwargs):
        """
        Initialize an Experiment instance, adding the necessary models and factors.

        Args:
            model: The base model.
            out_path: Output path for the CoreModel simulation.
            **kwargs: Additional parameters for CoreModel.
        """
        super().__init__(model, out_path, **kwargs)
        self.factor_names = []
        self.permutation = permutation
        # Add core experiment structure

        self.add_model(model_type=Models.Factorial.Experiment, adoptive_parent=Models.Core.Simulations)

        self.add_model(model_type=Models.Factorial.Factors, adoptive_parent=Models.Factorial.Experiment)

        if permutation:
            self.add_model(model_type=Models.Factorial.Permutation, adoptive_parent=Models.Factorial.Factors)

        # Add individual factors
        # self.add_model(model_type=Models.Factorial.Factor, adoptive_parent=Models.Factorial.Factors)

        # Move base simulation under the factorial experiment
        self.move_model(Models.Core.Simulation, Models.Factorial.Experiment, base_model_simulation, None)
        # pp = nexp.Simulations.FindInScope[Models.Factorial.Factor]()
        # pp.set_Specification("[Fertilise at sowing].Script.Amount = 0 to 100 step 20")
        self.save()

    def add_factor(self, specification: str, factor_name: str):
        # Add individual factors
        if factor_name in self.factor_names:
            raise ValueError(f"Factor {factor_name} already used")
        if self.permutation:
            parent_factor = Models.Factorial.Permutation
        else:
            parent_factor = Models.Factorial.Factors
        self.add_model(model_type=Models.Factorial.Factor, adoptive_parent=parent_factor, rename=factor_name)
        _added = self.Simulations.FindInScope[Models.Factorial.Factor](factor_name)
        _added.set_Specification(specification)
        self.save()
        self.factor_names.append(factor_name)


if __name__ == "__main__":
    from apsimNGpy.core.base_data import load_default_simulations

    path = load_default_simulations(simulations_object=False)
    experiment = Experiment(path, permutation=True)
    experiment.add_factor(specification="[Fertilise at sowing].Script.Amount = 0 to 200 step 20",
                          factor_name='Nitrogen')

    experiment.run()
    experiment.add_model(model_type=Models.Report, adoptive_parent=Models.Core.Simulation, rename='date')
    report = experiment.Simulations.FindInScope[Models.Report]('Report')
    Clock = Models.Clock()
    # report.set_VariableNames(Clock.Today)

    experiment.add_report_variable(commands='[Clock].Today as Date')
    experiment.preview_simulation()



from apsimNGpy.core.base_data import load_default_simulations, load_default_sensitivity_model
from apsimNGpy.core.apsim import ApsimModel
from apsimNGpy.tests.base_test import BaseTester, set_wd
import unittest


class TestCoreModel(BaseTester):
    def test_run(self):
        model = ApsimModel(model='Maize')
        model.run()
        self.test_ap_sim = model
        self.assertTrue(model.ran_ok)
        model.clean_up(db=True)

    def test_clone_model(self):
        model = ApsimModel(model='Maize')
        sim_name = 'clone_test'
        model.clone_model('Models.Core.Simulation', 'Simulation',
                          'Models.Core.Simulations', rename=sim_name)
        model.run()
        self.assertTrue(model.ran_ok)
        # inspect simulations
        sims = model.inspect_model(model_type='Models.Core.Simulation', fullpath=False)
        assert sim_name in sims, (f'{sim_name} is not among the current simulations, implying simulation was not '
                                  f'successful')
        model.clean_up(db=True)


# initialize the model
if __name__ == '__main__':
    unittest.main()
